/* The functions which calculate the energy, or cost of a particular
   ordering. */

#include <stdlib.h>

#include "order.h"
#include "seq.h"
#include "pages.h"
#include "cost.h"

#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#define MIN(a, b) (((a) < (b)) ? (a) : (b))

unsigned long
order_energy_seq (order_t * order, seq_t * seq)
{
  /* The basic algorithm is this:

     cost = 0
     Foreach arc <n1, n2, w> in the sequence graph
       p = number of unique pages touched by functions n1 and n2
       cost = cost + p*w

       Analysis: This algorithm tends to reduce the total number
       of pages that need to be in memory at any given time (in the
       window), but not the global minimum.
  */

  unsigned long cost;
  int page_size;
  list_t * curr;

  page_size = 4096;

  cost = 0;
  for (curr = seq->graph->arcs; curr != NULL; curr = curr->next)
    {
      int touched_pages;
      arc_t * curr_arc = (arc_t *) curr->data;
      func_t * func1, * func2;

      unsigned int high_page1, high_page2, low_page1, low_page2;

      func1 = order->func[order->func_idx[(int) curr_arc->n1->data]];
      func2 = order->func[order->func_idx[(int) curr_arc->n2->data]];

      low_page1 = func1->vma / page_size;
      high_page1 = (func1->vma + func1->size) / page_size;

      low_page2 = func2->vma / page_size;
      high_page2 = (func2->vma + func2->size) / page_size;

      /* Add up the total number of unique pages touched by the two
	 functions.  If the page sets are disjoint, then just add the
	 number of pages touched by each function.  Otherwise compute
	 the union. */
      if ((high_page2 < low_page1) || (high_page1 < low_page2))
	touched_pages = (high_page2 - low_page2 + 1) +
	  (high_page1 - low_page1 + 1);
      else
	touched_pages = MAX(high_page2, high_page1) -
	  MIN(low_page2, low_page1) + 1;

      cost += (touched_pages) * curr_arc->w;
    }

  return cost;
} /* order_energy_seq */

/*
 * Return the number of page faults generated by this ordering
 */
unsigned long
order_energy_pf (order_t * order, seq_t * seq)
{
  int touched_pages, i;
  pages_t * pages;
  list_t * curr;

  pages = pages_new(order->func[0]->vma,
		    order->func[order->num_funcs - 1]->vma +
		    order->func[order->num_funcs - 1]->size,
		    4096);

  for (curr = seq->graph->nodes; curr != NULL ; curr = curr->next)
    {
      node_t * node = (node_t *)curr->data;
      func_t * func;

      func = order->func[order->func_idx[(int) node->data]];

      page_span_inc(pages, func->vma, func->size, 1);
    }

  /* Count up all the touched pages */
  touched_pages = 0;
  for (i = 0 ; i < pages->num_pages; i ++)
    if (pages->page[i] != 0)
      touched_pages ++;

  pages_destroy(pages);

  return touched_pages;
} /* order_energy_pf */






