/* So how do we represent an ordering?  We want to be able to add
   holes and stuff.  Hmm.  So it cannot just be an order.

   We can represent orders like this:
   func1 = <offset from end of last function>
   func2 = <offset from end of last function>
   .
   .

   I think that is good.  So then what is a random step?
   aaaBBB.cccDDDD..EEEffffGGGGG

   Hmm.  For the traveling salesman algorithm, they used two types of
   steps:

   1. A section of the path is reversed
   2. A section of the path is removed and then placed in between
   two other elements of the path

   Perhaps I should have a bunch of diverse types of steps like that.

   1. Randomize the order of a section of the path
   2. Move a section of the path between two elements
   3. Swap two elements
   4. Increase the offset from the previous function.

   Hmm.  Ok, that seems like it might be ok.  So I will try that,
   except I will not implement the holes quite yet.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "order.h"
#include "seq.h"
#include "step.h"
#include "cost.h"
#include "anneal.h"

static order_t *
anneal(order_t * order, seq_t * seq)
{
  long curr_energy, new_energy, best_energy;
  order_t * new_order, * curr_order, * best_order;
  double temperature;
  int i;

  temperature = 0.3;

  curr_order = order_copy(order);

  curr_energy = order_energy_seq (curr_order, seq);
  printf("Starting energy: %ld\n", curr_energy);

  best_energy = curr_energy;
  best_order = order_copy(curr_order);

  for (i = 0 ; i < 1000 ; i ++)
    {
      /* Make 'new_order' be a random step away from 'curr_order' */
      new_order = order_copy(curr_order);
      order_step(new_order);
      
      new_energy = order_energy_seq (new_order, seq);

      if (metrop(curr_energy, new_energy, temperature))
	{
	  /* Take the step! */
	  order_destroy(curr_order);
	  curr_order = new_order;
	  curr_energy = new_energy;
	  printf("Stepped to %ld\n", curr_energy);

	  if (curr_energy < best_energy)
	    {
	      printf("This is a new minimum.\n");
	      best_energy = curr_energy;
	      order_destroy(best_order);
	      best_order = order_copy(curr_order);
	    }
	}
      else
	{
	  order_destroy(new_order);
	}

      temperature = temperature * 0.95;
    }

  order_destroy(curr_order);

  return best_order;
} /* anneal */

int
main(int argc, char ** argv)
{
  order_t * order, * new_order;
  seq_t * seq;
  
  if (argc < 3)
    {
      fprintf(stderr, "Usage: %s <executable> <sequence graph>\n",
	      *argv);
      exit(1);
    }

  srand(getpid());

  printf("Loading executable...\n");
  order = order_load(argv[1]);
  if (order == NULL)
    {
      fprintf(stderr, "Error loading %s\n", argv[1]);
      exit(1);
    }

  printf("Loading sequence graph...\n");
  seq = seq_load(argv[2], order);
  if (seq == NULL)
    {
      fprintf(stderr, "Error loading %s\n", argv[2]);
      exit(1);
    }

  fprintf(stderr, "Filtering out uncalled functions...\n");
  new_order = order_filter_seq(order, seq);
  order_destroy(order);
  order = new_order;

  fprintf(stderr, "Reloading sequence graph...\n");
  seq_destroy(seq, order);
  seq = seq_load(argv[2], order);

  printf("Annealing...\n");
  new_order = anneal (order, seq);

  printf("New order:\n");
  order_print(new_order);
  
  return 0;
} /* main */

