#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <bfd.h>

#include "order.h"
#include "symtab.h"

void
order_print(order_t * order)
{
  int i;

  for (i = 0 ; i < order->num_funcs; i ++)
    {
      fprintf(stderr, "Func[%d] == %s, vma: 0x%lx, size: %ld\n", i,
	      order->func[i]->name, order->func[i]->vma,
	      order->func[i]->size);
    }
} /* order_print */

order_t *
order_load (char * filename)
{
  asymbol ** symtab;
  int num_syms, i;
  order_t * order;
  bfd * abfd;

  abfd = bfd_openr(filename, NULL);
    if (abfd == NULL)
    {
      bfd_perror("bfd_openr");
      return NULL;
    }

  if (! bfd_check_format(abfd, bfd_object))
    {
      bfd_perror("bfd_check_format");
      bfd_close(abfd);
      return NULL;
    }

  symtab_load(abfd, &symtab, &num_syms);
  symtab_sort(symtab, num_syms);
  symtab_filter(&symtab, &num_syms, BSF_FUNCTION);

  order = (order_t *) malloc(sizeof(order_t));
  if (order == NULL)
    {
      fprintf(stderr, "core_load: Out of memory\n");
      bfd_close(abfd);
      free(symtab);
      return NULL;
    }

  order->num_funcs = num_syms;
  order->func = (func_t **) malloc(sizeof(func_t *) * order->num_funcs);
  order->func_idx = (unsigned int *)
    malloc(sizeof(unsigned int) * order->num_funcs);
  if (order->func == NULL || order->func_idx == NULL)
    {
      fprintf(stderr, "core_load: Out of memory\n");
      bfd_close(abfd);
      free(order);
      free(symtab);
      return NULL;
    }

  for (i = 0 ; i < order->num_funcs ; i++)
    {
      order->func[i] = (func_t *) malloc(sizeof(func_t));
      if (order->func[i] == NULL)
	{
	  fprintf(stderr, "core_load: Out of memory\n");
	  bfd_close(abfd);
	  free(symtab);
	  free(order->func);
	  free(order);
	  /* Oh well, we missed freeing some func's */
	  return NULL;
	}

      /* Keep track of an index which will always be used to reference
	 this function, no matter how much we rearrange things. */
      order->func_idx[i] = i;
      order->func[i]->idx = i;

      order->func[i]->name = strdup(symtab[i]->name);
      order->func[i]->vma = symtab[i]->value + symtab[i]->section->vma;
      order->func[i]->data = NULL;

      if ((i + 1) != order->num_funcs)
	order->func[i]->size =
	  (symtab[i + 1]->section->vma + symtab[i + 1]->value) -
	  (symtab[i]->section->vma + symtab[i]->value);
      else
	order->func[i]->size = symtab[i]->section->_raw_size -
	  symtab[i]->value;
    }

  free(symtab);
  bfd_close(abfd);

  return order;
} /* order_load */

func_t *
order_addr_to_func(order_t * order, unsigned long addr)
{
  int high_idx, low_idx, mid_idx;

  mid_idx = 0;
  for (high_idx = order->num_funcs, low_idx = 0; low_idx != high_idx;)
    {
      mid_idx = (high_idx + low_idx) / 2;

      if ((order->func[mid_idx]->vma <= addr) &&
	  (order->func[mid_idx]->vma + order->func[mid_idx]->size) >= addr)
	break;

      if (order->func[mid_idx]->vma > addr)
	high_idx = mid_idx;
      else
	low_idx = mid_idx + 1;
    }

  return order->func[mid_idx];
} /* order_addr_to_func */

order_t *
order_copy(order_t * src)
{
  order_t * dst;
  int i;

  dst = (order_t *) malloc(sizeof(order_t));
  if (dst == NULL)
    {
      fprintf(stderr, "order_copy: Out of memory\n");
      return NULL;
    }
    
  dst->num_funcs = src->num_funcs;
  dst->func = (func_t **) malloc(sizeof(func_t *) * dst->num_funcs);
  dst->func_idx = (unsigned int *) malloc(sizeof(func_t *) * dst->num_funcs);
  if (dst->func == NULL || dst->func_idx == NULL)
    {
      fprintf(stderr, "order_copy: Out of memory\n");
      return NULL;
    }

  for (i = 0 ; i < src->num_funcs ; i++)
    {
      dst->func[i] = (func_t *) malloc(sizeof(func_t));
      if (dst->func[i] == NULL)
	{
	  fprintf(stderr, "order_copy: Out of memory\n");
	  return NULL;
	}

      dst->func_idx[i] = src->func_idx[i];

      dst->func[i]->idx = src->func[i]->idx;
      dst->func[i]->vma = src->func[i]->vma;
      dst->func[i]->size = src->func[i]->size;
      dst->func[i]->name = strdup(src->func[i]->name);
    }

  return dst;
} /* order_copy */

void
order_destroy(order_t * order)
{
  order_destroy_funcs(order);
  free(order);
} /* order_destroy */

void
order_destroy_funcs(order_t * order)
{
  int i;

  for (i = 0 ; i < order->num_funcs ; i++)
    {
      free(order->func[i]->name);
      free(order->func[i]);
    }

  free(order->func);
  free(order->func_idx);
} /* order_destroy_funcs */

/*
 *
 * Randomize the order of the functions between start_func and
 * end_func, inclusive.  Does not adjust function VMAs.
 *
 */
void
order_span_randomize(order_t * order, int start_func, int end_func)
{
  int num_funcs, i;

  num_funcs = end_func - start_func + 1;

  /* Randomize the order of the functions in the specified section */
  for (i = num_funcs - 1 ; i > 0 ; i --)
    {
      unsigned int idx_a, idx_b;
      func_t * temp;
      int func;

      func = rand() % (i + 1);

      /* Swap functions 'i' and 'func'.  Also swap the indices into the
         func_idx array (this array provides the level of indirection
         which makes it possible to always find a function in constant
         time...) */
      idx_a = order->func[i + start_func]->idx;
      idx_b = order->func[func + start_func]->idx;
      order->func_idx[idx_a] = func + start_func;
      order->func_idx[idx_b] = i + start_func;

      temp = order->func[i + start_func];
      order->func[i + start_func] = order->func[func + start_func];
      order->func[func + start_func] = temp;
    }  
} /* order_span_randomize */

void
order_fix_vmas(order_t * order, unsigned long low_pc)
{
  unsigned long vma;
  int i;

  vma = low_pc;

  for (i = 0 ; i < order->num_funcs ; i ++)
    {
      order->func[i]->vma = vma;
      vma += order->func[i]->size;
    }
} /* order_fix_vmas */ 


/*
 *
 * Return a new order_t containing only those functions found
 * in seq.
 *
 */
order_t * 
order_filter_seq (order_t * order, seq_t * seq)
{
  unsigned int curr_func;
  order_t * new;
  int i;

  new = (order_t *) malloc(sizeof(order_t));
  if (new == NULL)
    {
      fprintf(stderr, "order_filter_seq: Out of memory\n");
      exit(1);
    }

  /* Allocate more space than we need.. */
  new->func = (func_t **) malloc(sizeof(func_t *) * order->num_funcs);
  new->func_idx = (unsigned int *)
    malloc(sizeof(unsigned int) * order->num_funcs);
  if (new->func == NULL || new->func_idx == NULL)
    {
      fprintf(stderr, "order_filter_seq: Out of memory\n");
      exit(1);
    }

  curr_func = 0;
  for (i = 0 ; i < order->num_funcs; i ++)
    {
      if (seq_func_present(seq, order->func[i]))
	{
	  new->func[curr_func] = (func_t *) malloc(sizeof(func_t));
	  if (new->func[curr_func] == NULL)
	    {
	      fprintf(stderr, "order_filter_seq: Out of memory\n");
	      exit(1);
	    }

	  new->func[curr_func]->vma = order->func[i]->vma;
	  new->func[curr_func]->size = order->func[i]->size;
	  new->func[curr_func]->name = strdup(order->func[i]->name);

	  new->func[curr_func]->idx = curr_func;
	  new->func_idx[curr_func] = curr_func;
	  curr_func ++;
	}
    }

  new->num_funcs = curr_func;

  printf("%d funcs remain\n", new->num_funcs);

  return new;
} /* order_filter_seq */



