#include <stdio.h>
#include <stdlib.h>

#include <gmon_out.h>

#include "order.h"
#include "seq.h"

/* NOTE: The 'order' must be sorted ! */
seq_t *
seq_load(char * filename, order_t * order)
{
  struct gmon_cg_arc_record raw_arc;
  unsigned char tag;
  seq_t * seq;
  FILE * f;

  f = fopen(filename, "r");
  if (f == NULL)
    {
      fprintf(stderr, "seq_load: Could not open %s\n", filename);
      perror("fopen");
      return NULL;
    }

  seq = (seq_t *) malloc(sizeof(seq_t));
  if (seq == NULL)
    {
      fprintf(stderr, "seq_load: Out of memory\n");
      fclose(f);
      return NULL;
    }

  seq->graph = graph_new();
  
  /* The data in the sequence file is stored just like a gmon.out
     callgraph. */
  while (fread (&tag, sizeof(tag), 1, f) == 1)
    {
      if (tag != GMON_TAG_CG_ARC)
	{
	  fprintf(stderr, "seq_load: Invalid sequence graph: %s\n", filename);
	  free(seq);
	  return NULL;
	}

      if (fread(&raw_arc, sizeof(raw_arc), 1, f) != 1)
	{
	  fprintf(stderr, "seq_load: Corrupted file: %s\n", filename);
	  free(seq);
	  return NULL;
	}

      /* Add this arc to the sequence graph */
      seq_arc(seq, order,
	      * (unsigned long *) raw_arc.from_pc,
	      * (unsigned long *) raw_arc.self_pc,
	      * (unsigned long *) raw_arc.count);
    }

  graph_arc_sort(seq->graph);

#if 0
  {
    list_t * curr;
    for (curr = seq->graph->arcs; curr != NULL ; curr = curr->next)
      {
	arc_t * arc = (arc_t *)  curr->data;
	printf("Arc: <%s, %s, %ld>\n", ((func_t *)arc->n1->data)->name,
	   ((func_t *)arc->n2->data)->name, arc->w);
      }
  }
#endif

  fclose(f);
  return seq;
} /* seq_load */

void
seq_destroy(seq_t * seq, order_t * order)
{
  int i;

  graph_destroy(seq->graph);

  for (i = 0 ; i < order->num_funcs ; i++)
    order->func[i]->data = NULL;
  
  free(seq);
} /* seq_destroy */

void
seq_arc(seq_t * seq, order_t * order, unsigned long addr1,
	unsigned long addr2, unsigned long weight)
{
  func_t * func1, * func2;
  node_t * node1, * node2;

  /* Find the functions associated with these addresses */
  func1 = order_addr_to_func(order, addr1);
  func2 = order_addr_to_func(order, addr2);

  /* Now find the graph nodes associated with these functions */
  node1 = (node_t *) func1->data;
  if (node1 == NULL)
    {
      /* Store the function's index into the func_idx array in the node's
	 private data element for later retrieval. */
      node1 = node_new(0, (void *) func1->idx);
      func1->data = (void *) node1;
      graph_node_add(seq->graph, node1);
    }

  node2 = (node_t *) func2->data;
  if (node2 == NULL)
    {
      node2 = node_new(0, (void *) func2->idx);
      func2->data = (void *) node2;
      graph_node_add(seq->graph, node2);
    }

  /* Now increase the weight of the arc between these two nodes in the
     graph. */
  graph_arc_inc(seq->graph, node1, node2, weight);
} /* seq_arc */


/*
 * Returns true if the specified function is present in the sequence
 * graph, and false otherwise.
 */
int
seq_func_present(seq_t * seq, func_t * func)
{
  list_t * curr;

  for (curr = seq->graph->nodes ; curr != NULL ; curr = curr->next)
    {
      node_t * node;
      int idx;

      node = (node_t *) curr->data;
      idx = (int) node->data;

      if (idx == func->idx)
	return 1;
    }

  return 0;
} /* seq_func_present */

/*
 * Make the private 'data' element of each node in the graph point
 * to the name of the function in the order, not the function index.
 *
 */
void
seq_resolve_names(seq_t * seq, order_t * order)
{
  list_t * curr;

  for (curr = seq->graph->nodes ; curr != NULL ; curr = curr->next)
    {
      node_t * node = (node_t *) curr->data;

      node->data = order->func[order->func_idx[(int) node->data]]->name;
    }
} /* seq_resolve_names */
