#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>

#include "order.h"
#include "func.h"
#include "pages.h"

#include "sim.h"

sim_t *
sim_load(char * filename, order_t * order)
{
  struct stat stat_buf;
  int call_num, i;
  sim_t * sim;
  unsigned long vma;
  char * map;
  FILE * f;

  f = fopen(filename, "r");
  if (f == NULL)
    {
      fprintf(stderr, "sim_load: Unable to open sim file %s\n", filename);
      perror("fopen");
      return NULL;
    }

  if (fstat(fileno(f), &stat_buf) < 0)
    {
      perror("fstat");
      fprintf(stderr, "sim_load: Could not determine size of sim file %s\n",
	      filename);
      return NULL;
    }

  sim = (sim_t *) malloc(sizeof(sim_t));
  if (sim == NULL)
    {
      fprintf(stderr, "sim_load: Out of memory\n");
      return NULL;
    }

  sim->num_calls = stat_buf.st_size / sizeof(unsigned long);
  sim->call = (func_t **) malloc(sizeof(func_t *) * sim->num_calls);
  if (sim->call == NULL)
    {
      fprintf(stderr, "sim_load: Out of memory\n");
      free(sim);
      fclose(f);
      return NULL;
    }

  map = (char *) mmap(NULL, stat_buf.st_size, PROT_READ, MAP_SHARED,
		      fileno(f), 0);
  if (map == MAP_FAILED)
    {
      fprintf(stderr, "sim_load: Could not mmap sim file!\n");
      free(sim->call);
      free(sim);
      fclose(f);
      return NULL;
    }

  for (call_num = 0 ; call_num < sim->num_calls; call_num++)
    {
      int high_idx, low_idx;

      vma = *(unsigned long *) (map + call_num*4);

      /* Locate this VMA in the order */
      i = 0;
      for (high_idx = order->num_funcs, low_idx = 0; low_idx != high_idx;)
	{
	  i = (high_idx + low_idx) / 2;

	  if ((order->func[i]->vma <= vma) &&
	      (order->func[i]->vma + order->func[i]->size) >= vma)
	    break;

	  if (order->func[i]->vma > vma)
	    high_idx = i;
	  else
	    low_idx = i + 1;
	}

      sim->call[call_num] = order->func[i];
    }

  if (munmap(map, stat_buf.st_size) < 0)
    perror("munmap");

  fclose(f);

  return sim;
} /* sim_load */

#if 0
int
sim_run(order_t * order, sim_t * sim, int * total_pf, int * avg_set_size)
{
  pages_t * overall, * working_set;
  unsigned long i;
  int window;

  /* We use 'overall' to determine the total number of page faults generated.
     We use 'working_set' to determine the average working set size. */
  overall = pages_new(order->func[0]->vma,
		      order->func[order->num_funcs - 1]->vma +
		      order->func[order->num_funcs - 1]->size,
		      4096);

  working_set = pages_new(order->func[0]->vma,
			  order->func[order->num_funcs - 1]->vma +
			  order->func[order->num_funcs - 1]->size,
			  4096);


  /* First determine the total number of page faults */
  for (i = 0 ; i < sim->num_calls; i++)
    {
      unsigned long addr;

      for (addr = sim->call[i]->vma ;
	   addr < (sim->call[i]->vma + sim->call[i]->size);
	   addr += (overall->page_size - (addr % overall->page_size)))
	page_inc(overall, addr);
    }

  *total_pf = 0;
  for (i = 0 ; i < overall->num_pages ; i ++)
    {
      if (overall->page[i] != 0)
	(*total_pf) ++;
    }

  /* Now find the average working set size.  Define the working set as
     the number of pages that need to be resident in a call window of
     'window' functions */
  *avg_set_size = 0;
  window = 10;
  for (i = 0 ; i < sim->num_calls ; i ++)
    {
      unsigned long j;
      int working_set_size;
      pages_clear(working_set);
      for (j = i ; (j < sim->num_calls) && (j < (i + window)); j ++)
	{
	  unsigned long addr;

	  for (addr = sim->call[j]->vma ;
	       addr < (sim->call[j]->vma + sim->call[j]->size);
	       addr += (overall->page_size - (addr % overall->page_size)))
	    page_inc(overall, addr);
	}

      working_set_size = 0;
      for (j = 0 ; j < overall->num_pages ; j ++)
	{
	  if (overall->page[j] != 0)
	    working_set_size ++;
	}
    }

  return 0;
} /* sim_run */

#endif
