#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/gmon.h>
#include <sys/uio.h>
#include <sys/gmon_out.h>


/* Routines to store a graph of arcs between functions in memory.
   These are done here in a generalized way so that they can be reused
   in other profiling modes. */

typedef struct gmon_graph_struct {
  /* "arcs" actually holds the second half of the arc,
     the first half being stored in "nodes" */
  struct gmon_graph_arc_struct * arcs;

  /* All the nodes.  This is an array of indices which reference
     members of the "arcs" vector above.  To find the arc between a
     and b, you find a in nodes, and then follow the linked list of
     "arcs" until you reach b. */
  u_long * nodes;

  u_long num_nodes, num_arcs;
  u_long next_arc;
} gmon_graph_t;

typedef struct gmon_graph_arc_struct {
  u_long addr;
  u_long count;

  /* Next index */
  u_long next;
} gmon_graph_arc_t;

gmon_graph_t *
gmon_graph_new(u_long text_size)
{
  gmon_graph_t * graph;

  graph = malloc(sizeof(gmon_graph_t));
  if (graph == NULL)
    {
      fprintf(stderr, "gmon_graph_new: Out of memory\n");
      return NULL;
    }

  graph->num_arcs = text_size * ARCDENSITY / 100;
  graph->num_nodes = text_size / HASHFRACTION;

  graph->nodes = malloc(sizeof(u_long) * graph->num_nodes);
  graph->arcs = malloc(sizeof(gmon_graph_arc_t) * graph->num_arcs);
  if (graph->arcs == NULL || graph->nodes == NULL)
    {
      fprintf(stderr, "gmon_graph_new: Out of memory\n");
      return NULL;
    }

  memset(graph->nodes, 0, sizeof(u_long) * graph->num_nodes);
  memset(graph->arcs, 0, sizeof(gmon_graph_arc_t) * graph->num_arcs);

  /* We haven't used any of the arcs up yet */
  graph->next_arc = 0;

  return graph;
} /* gmon_graph_new */

/* Make an arc between addr1 and addr2 */
void
gmon_graph_arc(gmon_graph_t * graph, u_long addr1, u_long addr2)
{
  u_long curr_idx, prev_idx;

  /* Find the arc entry */
  prev_idx = 0;
  curr_idx = graph->nodes[addr1 / HASHFRACTION];
  while ( (curr_idx != 0) && (graph->arcs[curr_idx].addr != addr2))
    {
      prev_idx = curr_idx;
      curr_idx = graph->arcs[curr_idx].next;
    }

  /* No entry in the graph; add a new entry at the front */
  if (curr_idx == 0)
    {
      /* Allocate a new arc slot */
      if (graph->next_arc == graph->num_arcs)
	{
	  /* We ran out of slots */
	  fprintf(stderr, "gmon_graph_arc: No more arc slots!\n");
	  exit(1);
	}
      curr_idx = graph->next_arc;
      graph->next_arc ++;

      graph->arcs[curr_idx].addr = addr2;
      graph->arcs[curr_idx].count = 1;

      /* Add it to the front of the list */
      graph->arcs[curr_idx].next = graph->nodes[addr1 / HASHFRACTION];
      graph->nodes[addr1 / HASHFRACTION] = curr_idx;
    }
  else
    {
      /* We found the node; increment its arc count and advance it to the
	 front of the list if it is not already there. */
      graph->arcs[curr_idx].count ++;

      if (prev_idx != 0)
	{
	  graph->arcs[prev_idx].next = graph->arcs[curr_idx].next;
	  graph->arcs[curr_idx].next = graph->nodes[addr1 / HASHFRACTION];
	  graph->nodes[addr1 / HASHFRACTION] = curr_idx;
	}
    }
  
} /* gmon_graph_arc */

int
gmon_graph_write_gmon(gmon_graph_t * graph, u_long lowpc, int fd)
{
  u_char tag = GMON_TAG_CG_ARC;
  struct gmon_cg_arc_record raw_arc[4]
    __attribute__ ((aligned (__alignof__ (char*))));
  int arc_index, node_index;
  u_long frompc;

  struct iovec iov[8] =
    {
      { &tag, sizeof (tag) },
      { &raw_arc[0], sizeof (struct gmon_cg_arc_record) },
      { &tag, sizeof (tag) },
      { &raw_arc[1], sizeof (struct gmon_cg_arc_record) },
      { &tag, sizeof (tag) },
      { &raw_arc[2], sizeof (struct gmon_cg_arc_record) },
      { &tag, sizeof (tag) },
      { &raw_arc[3], sizeof (struct gmon_cg_arc_record) },
    };
  int nfilled = 0;

  for (node_index = 0; node_index < graph->num_nodes; ++node_index)
    {
      if (graph->nodes[node_index] == 0)
	continue;

      frompc = lowpc + node_index * HASHFRACTION;
      for (arc_index = graph->nodes[node_index];
	   arc_index != 0;
	   arc_index = graph->arcs[arc_index].next)
	{
	  if (nfilled > 3)
	    {
	      __writev (fd, iov, 2 * nfilled);
	      nfilled = 0;
	    }
	  *(char **) raw_arc[nfilled].from_pc = (char *)frompc;
	  *(char **) raw_arc[nfilled].self_pc =
	    (char *) graph->arcs[arc_index].addr;
	  *(int *) raw_arc[nfilled].count = graph->arcs[arc_index].count;
	  ++nfilled;
	}
    }
  __writev (fd, iov, 2 * nfilled);
  return 0;
} /* gmon_graph_write_gmon */

