#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "seq_graph.h"

/* Make a function affinity graph based on call sequences */


/* The working set is the last N function addresses which we know were
   accessed.  This is like a sliding window of function addresses.
   Adjacent addresses in this window are never identical so that
   recursive calls do not mess up the data.

   The actual window size used during the run is configurable via the
   environment variable GPROF_WINDOW_SIZE. */

u_long __seq_graph_window[__SEQ_GRAPH_WINDOW_SIZE];
u_short __seq_graph_winpos; /* Where we put the last element in the
			       window.  This value wraps to create the
			       sliding window effect. */
u_short __seq_graph_winsize = 10;

u_long __seq_graph_lowpc;

/* The graph.  It's big. */
u_long ** __seq_graph;
u_long __seq_graph_size;

inline void
gmon_seq_arc(u_long pc1, u_long pc2)
{
  __seq_graph[(pc1 - __seq_graph_lowpc) / __SEQ_GRAPH_ARC_DENSITY]
    [(pc2 - __seq_graph_lowpc) / __SEQ_GRAPH_ARC_DENSITY] ++;
} /* gmon_seq_arc */

int
gmon_seq_init(u_long lowpc, u_long highpc)
{
  u_long text_size;

  text_size = highpc - lowpc;

  fprintf(stderr, "gmon_seq_init(lowpc == %ld, highpc == %ld)\n",
	  lowpc, highpc);
  fprintf(stderr, "gmon_seq_init: text_size == %uld\n", text_size);
  

  __seq_graph_size = (text_size / 256) * (text_size / 256) * sizeof(u_long);
  __seq_graph = (u_long **) malloc(__seq_graph_size);
  if (__seq_graph == NULL)
    {
      fprintf(stderr, "gmon_seq_init: Out of memory\n");
      /* Returning 0 will disable our functionality */
      return 0;
    }

  fprintf(stderr, "gmon_seq_init: Allocated %uld bytes for graph\n",
	  __seq_graph_size);
  
  memset((void *)__seq_graph, 0, __seq_graph_size);

  fprintf(stderr, "gmon_seq_init: Cleared graph\n");
  
  __seq_graph_lowpc = lowpc;

  if (getenv("GPROF_WINDOW_SIZE") != NULL)
    {
      __seq_graph_winsize = atoi(getenv("GPROF_WINDOW_SIZE"));
      if (__seq_graph_winsize > __SEQ_GRAPH_WINDOW_SIZE)
	__seq_graph_winsize = __SEQ_GRAPH_WINDOW_SIZE;
    }
  
      

  return 1;
} /* gmon_seq_init */

/* This function is called when a function is accessed, either on
   function call or function return. */
void
gmon_seq_access(u_long pc)
{
  /* Make sure that this is not the same function address which is
     already in there so that we don't create lots of duplicates.
     FIXME: is this really what we want to do? */
  if (pc == __seq_graph_window[__seq_graph_winpos])
    return;

  fprintf(stderr, "gmon_seq_access\n");
  
  /* Move on to the next window position */
  __seq_graph_winpos++;
  if (__seq_graph_winpos == __seq_graph_winsize)
    __seq_graph_winpos = 0;

  __seq_graph_window[__seq_graph_winpos] = pc;

  /* The address is now added to the window.  Now, create arcs between
     all of the functions listed in the window. */
  gmon_seq_arcs();
  
} /* gmon_seq_access */

void
gmon_seq_arcs(void)
{
  int i, j;

  /* Create arcs between all of the functions listed in the window. */
  for (i = 0 ; i < __seq_graph_winsize ; i++)
    for ( j = i + 1 ; j < __seq_graph_winsize ; j ++)
      {
	/* Create an arc between these two entries */
	if ((__seq_graph_window[i] != 0) && (__seq_graph_window[j] != 0))
	  if (__seq_graph_window[i] != __seq_graph_window[j])
	    gmon_seq_arc(__seq_graph_window[i], __seq_graph_window[j]);
      }

} /* gmon_seq_arc */




void
gmon_seq_write(void)
{
  int fd;

  fd = creat("./gmon.seq.out", 0600);
  if (fd < 0)
    {
      fprintf(stderr, "gmon_seq_write: Could not open file\n");
      perror("creat");
      return;
    }

  if (write(fd, (char *) &__seq_graph_size, sizeof(__seq_graph_size)) < 0)
    {
      fprintf(stderr, "gmon_seq_write: Could not write graph size to file!\n");
      perror("write");
    }

  if (write(fd, (char *) __seq_graph, __seq_graph_size) < __seq_graph_size)
    {
      fprintf(stderr, "gmon_seq_write: Could not write data to file!\n");
      perror("write");
      return;
    }

  close(fd);
    
} /* gmon_seq_write */

