#include <stdio.h>
#include <stdlib.h>
#include <bfd.h>

int verbose = 1;

static int symtab_sort_compare(const void * a, const void * b);

static void
symbol_print(asymbol * sym)
{

  if (strlen(sym->name) == 0)
    printf("\"\" ");
  else
    printf("%s ", sym->name);

  if (sym->flags & BSF_FUNCTION)
    printf("func ");

  if (sym->flags & BSF_DEBUGGING)
    printf("debug ");

  if (sym->flags & BSF_LOCAL)
    printf("local ");

  if (sym->flags & BSF_GLOBAL)
    printf("global ");

  if (sym->flags & BSF_WEAK)
    printf("weak ");

  if (sym->flags & BSF_CONSTRUCTOR)
    printf("constructor-sect ");

  if (sym->flags & BSF_WARNING)
    printf("warning ");

  if (sym->flags & BSF_INDIRECT)
    printf("indirect ");

  if (sym->flags & BSF_FILE)
    printf("file ");

  if (sym->flags & BSF_DYNAMIC)
    printf("dynamic ");

  if (sym->flags & BSF_OBJECT)
    printf("object ");

  if (sym->flags & BSF_SECTION_SYM)
    printf("section ");

  printf("0x%lx (0x%lx) ", (sym->value + sym->section->vma), sym->value);

  printf("%s ", sym->section->name);
  printf("\n");

} /* symbol_print */

static void
symtab_print(asymbol ** symtab, int num_syms)
{
  int i;

  for (i = 0 ; i < num_syms ; i ++)
    symbol_print(symtab[i]);

} /* symtab_print */

/* Return the function symbol in which a vma is found.  NULL if not found.
   The symbol table must be sorted first and must only contain function
   symbols. */
static asymbol *
symtab_vma_to_sym(asymbol ** symtab, int num_syms, bfd_vma vma)
{
  static bfd_vma vma_cache = 0;
  static asymbol * sym_cache = NULL;
  int i;

  if (vma == vma_cache)
    return sym_cache;
  
  for (i = 0 ; i < num_syms ; i++)
    {
      if ((symtab[i]->value + symtab[i]->section->vma) > vma)
	{
	  if (i == 0)
	    return NULL;

	  vma_cache = vma;
	  sym_cache = symtab[i - 1];
	  return sym_cache;
	}
    }

  return NULL;
} /* symtab_vma_to_sym */

/*
 * Remove any symbols which do not have 'flags' set
 */
static void
symtab_filter(asymbol *** symtab, int * num_syms, flagword flags)
{
  asymbol ** new_symtab;
  int new_num_syms, i, j;
  
  if (! flags)
    return;

  /* First find out how many symbols match the flags so that we know
     how much space to allocate */
  new_num_syms = 0;
  for (i = 0 ; i < *num_syms ; i++)
    {
      if ((*symtab)[i]->flags & flags)
	new_num_syms++;
    }

  if (new_num_syms == 0)
    new_symtab = NULL;
  else
    {
      new_symtab = (asymbol **) malloc(new_num_syms * sizeof(asymbol *));
      if (new_symtab == NULL)
	{
	  fprintf(stderr, "symtab_filter: Could not allocate %d bytes "
		  "for new symtab\n", new_num_syms * sizeof(asymbol *));
	  exit(1);
	}
    }

  /* Now copy all the matching symbols into place. */
  for (j = 0, i = 0 ; i < *num_syms ; i ++)
    {
      if ((*symtab)[i]->flags & flags)
	{
	  new_symtab[j] = (*symtab)[i];
	  j++;
	}
    }

  free (*symtab);
  *num_syms = new_num_syms;
  *symtab = new_symtab;
} /* symtab_filter */

/*
 * Sort the symbol table by symbol address
 */
static void
symtab_sort(asymbol ** symtab, int num_syms)
{
  qsort((void *) symtab, num_syms, sizeof(asymbol *), symtab_sort_compare);
} /* symtab_sort */

static int
symtab_sort_compare(const void * a, const void * b)
{
  asymbol * sym_a = * (asymbol **) a;
  asymbol * sym_b = * (asymbol **) b;

  bfd_vma val_a, val_b;

  val_a = sym_a->value + sym_a->section->vma;
  val_b = sym_b->value + sym_b->section->vma;

  if (val_a < val_b)
    return -1;

  if (val_a == val_b)
    return 0;

  return 1;
} /* symtab_sort_compare */

static void
symtab_load(bfd * abfd, asymbol *** symtab, int * num_syms)
{
  long size;

  size = bfd_get_symtab_upper_bound(abfd);
  if (size == -1)
    {
      bfd_perror("bfd_get_symtab_upper_bound");
      exit(1);
    }

  /* Allocate space for the vector of pointers to symbols */
  *symtab = (asymbol **) malloc(size);
  if (*symtab == NULL)
    {
      fprintf(stderr, "Could not allocate %ld bytes for the symbol table\n",
	      size);
      exit(1);
    }

  *num_syms = bfd_canonicalize_symtab (abfd, *symtab);
  if (*num_syms < 0)
    {
      bfd_perror("bfd_canonicalize_symtab");
      exit(1);
    }

  if (verbose)
    printf("symtab_load: Loaded %d symbols\n", *num_syms);

  if (verbose)
    symtab_print(*symtab, *num_syms);
  
} /* symtab_load */

int
main(int argc, char ** argv)
{
  bfd * abfd;
  asymbol ** symtab;
  asymbol * sym;
  int num_syms;
  bfd_vma addr;
  char line[1024];
  
  if (argc < 2)
    {
      fprintf(stderr, "Usage: %s <file>\n", *argv);
      exit(1);
    }

  bfd_init();

  abfd = bfd_openr(argv[1], NULL);
  if (abfd == NULL)
    {
      bfd_perror("bfd_openr");
      exit(1);
    }

  if (! bfd_check_format(abfd, bfd_object))
    {
      bfd_perror("bfd_check_format");
      exit(1);
    }

  symtab_load(abfd, &symtab, &num_syms);
  symtab_sort(symtab, num_syms);
  symtab_filter(&symtab, &num_syms, BSF_FUNCTION);
  printf("\n\n");
  symtab_print(symtab, num_syms);

  while (!feof(stdin))
    {
      if (fgets(line, sizeof(line), stdin) == NULL)
	exit(0);

      addr = strtol(line, NULL, 16);
      sym = symtab_vma_to_sym(symtab, num_syms, addr);
      if (sym != NULL)
	printf("%s\n", sym->name);
    }
  
  return 0;
} /* main */
