#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "graph.h"

#define xfree(x) { free(x); x = NULL; }

/*
 *
 * Lists
 *
 */

list_t *
list_new(void * data)
{
  list_t * list;

  list = malloc(sizeof(list_t));
  if (list == NULL)
    {
      fprintf(stderr, "list_new: Couldn't allocate space for new list\n");
      perror("malloc");
      exit(1);
    }

  list->data = data;
  list->next = NULL;

  return list;
} /* list_new */

void
list_delete(list_t * list)
{
  list_t * curr, * temp;

  assert(list != NULL);

  curr = list;
  while (curr != NULL)
    {
      temp = curr;
      curr = curr->next;
      xfree (temp);
    }

} /* list_delete */ 

list_t *
list_remove_element(list_t * list, void * element)
{
  list_t * curr, * prev, * temp;

  assert(list != NULL);

  curr = list;
  prev = NULL;
  while (curr != NULL)
    {
      if (curr->data == element)
	{
	  /* If this is the first element in the list... */
	  if (prev == NULL)
	    {
	      temp = curr->next;
	      xfree(curr);
	      return temp;
	    }

	  /* Otherwise, link around this element */
	  prev->next = curr->next;
	  xfree (curr);
	  return list;
	}

      prev = curr;
      curr = curr->next;
    }
  
  return list;
} /* list_remove_element */

int
list_element_present_p(list_t * list, void * element)
{
  list_t * curr;

  for (curr = list ; curr != NULL ; curr = curr->next)
    {
      if (curr->data == element)
	return TRUE;
    }

  return FALSE;
} /* list_element_present_p */

/* Add a link containing 'data' to the list */
list_t *
list_add(list_t * list, void * data)
{
  list_t * new;

#if 0
  if (list_element_present_p(list, data))
    {
      return list;
    }
#endif

  new = list_new(data);
  new->next = list;
  return new;
} /* list_add */

/*
 *
 * Nodes
 *
 */

node_t *
node_new (weight_t w, void * data)
{
  node_t * node;

  node = malloc(sizeof(node_t));
  if (node == NULL)
    {
      fprintf(stderr, "node_new: Couldn't allocate space for new node\n");
      perror("malloc");
      exit(1);
    }

  node->w = w;
  node->data = data;
  node->arcs = NULL;

  return node;
} /* node_new */

/*
 *
 * Arcs
 *
 */

arc_t *
arc_new (node_t * n1, node_t * n2, weight_t w, void * data)
{
  arc_t * arc;

  assert(n1 != NULL);
  assert(n2 != NULL);

  arc = malloc(sizeof(arc_t));
  if (arc == NULL)
    {
      fprintf(stderr, "arc_new: Couldn't allocate space for new arc\n");
      perror("malloc");
      exit(1);
    }

  arc->w = w;
  arc->n1 = n1;
  arc->n2 = n2;
  arc->data = data;

  /* Add this arc to each node's list of arcs */
  n1->arcs = list_add(n1->arcs, arc);
  n2->arcs = list_add(n2->arcs, arc);

  return arc;
} /* arc_new */

void
arc_delete(arc_t * arc)
{
  assert (arc != NULL);

  arc->n1->arcs = list_remove_element(arc->n1->arcs, arc);
  arc->n2->arcs = list_remove_element(arc->n2->arcs, arc);

  xfree (arc);

} /* arc_delete */

/*
 *
 * Graphs
 *
 */

graph_t * 
graph_new(void)

{
  graph_t * graph;

  graph = malloc(sizeof(graph_t));
  if (graph == NULL)
    {
      fprintf(stderr, "graph_new: Could not allocate space for new graph\n");
      perror("malloc");
      exit(1);
    }

  graph->arcs = NULL;
  graph->nodes = NULL;

  return graph;
} /* graph_new */

void
graph_destroy(graph_t * graph)
{
  list_t * curr;

  /* Destroy all the arcs */
  curr = graph->arcs;
  while (curr != NULL)
    {
      list_t * next;
      next = curr->next;

      arc_delete((arc_t *)curr->data);
      free(curr);

      curr = next;
    }

  /* Destroy all the nodes */
  curr = graph->nodes;
  while (curr != NULL)
    {
      list_t * next;
      next = curr->next;

      free((node_t *)curr->data);
      free(curr);

      curr = next;
    }

  free(graph);

} /* graph_destroy */

void
graph_node_add(graph_t * graph, node_t * node)
{

  assert(graph != NULL);
  assert(node != NULL);

  graph->nodes = list_add(graph->nodes, node);
} /* graph_node_add */

void
graph_arc_add(graph_t * graph, arc_t * arc)
{

  assert(graph != NULL);
  assert(arc != NULL);

  if (arc->n1 == arc->n2)
    return;

  /* Make sure that each node of the arc is in the graph */
  /* FIXME */
  /*
    graph_node_add(graph, arc->n1);
    graph_node_add(graph, arc->n2);
  */

  /* We insert the arc into the graph with no regards for keeping the
     list of arcs sorted.  You must sort them with graph_arcs_sort(). */
  graph->arcs = list_add(graph->arcs, arc);
} /* graph_arc_add */

void
graph_arc_new(graph_t * graph, node_t * n1, node_t * n2, weight_t w,
	      void * data)
{
  arc_t * arc;

  assert(graph != NULL);
  assert(n1 != NULL);
  assert(n2 != NULL);

  if (n1 == n2)
    return;
  
  arc = arc_new(n1, n2, w, data);
  graph_arc_add(graph, arc);
} /* graph_arc_new */

/* Return the arc between the two specified nodes or NULL if no such
   arc currently exists */
arc_t *
graph_arc_get(graph_t * graph, node_t * n1, node_t * n2)
{
  list_t * curr;
  arc_t * arc;
#if 0
  assert(graph != NULL);
  assert(n1 != NULL);
  assert(n2 != NULL);
#endif

  for (curr = n1->arcs ; curr != NULL ; curr = curr->next)
    {
      arc = (arc_t *) curr->data;

      if ((arc->n1 == n2) || (arc->n2 == n2))
	return arc;
    }

  return NULL;
} /* graph_arc_get */

static int
graph_arc_sort_compare(const void * a, const void * b)
{
  arc_t * arc_a, * arc_b;

  arc_a = *(arc_t **) a;
  arc_b = *(arc_t **) b;

  if (arc_a->w > arc_b->w)
    return 1;
  if (arc_a->w < arc_b->w)
    return -1;

  return 0;
} /* graph_arc_sort_compare */

/* Sort the arcs in the graph */
void
graph_arc_sort(graph_t * graph)
{
  long num_arcs = 0, curr_arc;
  list_t * curr;
  arc_t ** arc_buff;

  /* Find out how many arcs there are and allocate a buffer to hold
     pointers to them all */
  for (curr = graph->arcs ; curr != NULL ; curr = curr->next)
    num_arcs ++;

  arc_buff = (arc_t **) malloc(sizeof(arc_t *) * num_arcs);
  if (arc_buff == NULL)
    {
      fprintf(stderr, "Could not allocate space to sort arcs\n");
      exit(1);
    }

  /* Now copy all the arc pointers into the buffer */
  curr_arc = 0;
  for (curr = graph->arcs ; curr != NULL ; curr = curr->next)
    {
      arc_buff[curr_arc]= (arc_t *) curr->data;
      curr_arc ++;
    }

  /* Deallocate the old list */
  list_delete(graph->arcs);
  graph->arcs = NULL;

  /* Sort the buffer */
  qsort((void *) arc_buff, num_arcs, sizeof(arc_t *),
	graph_arc_sort_compare);

  /* Make a new list from the sorted buffer */
  for (curr_arc = 0 ; curr_arc < num_arcs ; curr_arc ++)
    {
      list_t * new;

      new = list_new(arc_buff[curr_arc]);

      new->next = graph->arcs;
      graph->arcs = new;
    }
} /* graph_arc_sort */

/* Increment the weight of the arc between n1 and n2 by 'step,' or create it
   with weight 'step' if it doesn't exist. */
void
graph_arc_inc(graph_t * graph, node_t * n1, node_t * n2, weight_t step)
{
  arc_t * arc;

  assert(graph != NULL);
  assert(n1 != NULL);
  assert(n2 != NULL);

  if (n1 == n2)
    return;

  arc = graph_arc_get(graph, n1, n2);
  if (arc == NULL)
    {
      graph_arc_new(graph, n1, n2, step, NULL);
      return;
    }

  arc->w += step;

} /* graph_arc_inc */

/* Set the weight of the arc between n1 and n2 to 'weight.'  Create the
   arc if it doesn't exist */
void
graph_arc_set(graph_t * graph, node_t * n1, node_t * n2, int weight)
{
  arc_t * arc;

  assert(graph != NULL);
  assert(n1 != NULL);
  assert(n2 != NULL);

  if (n1 == n2)
    return;

  arc = graph_arc_get(graph, n1, n2);
  if (arc == NULL)
    {
      graph_arc_new(graph, n1, n2, weight, NULL);
      return;
    }

  arc->w = weight;
} /* graph_arc_set */

void
graph_arc_delete(graph_t * graph, arc_t * arc)
{
  list_t * curr, * prev;
  assert(graph != NULL);
  assert(arc != NULL);

#if 0
  printf("graph_arc_delete(%s, %s)\n",
	 (char *) arc->n1->data, (char *) arc->n2->data);
#endif

  /* Remove it from the graph's list of arcs */
  curr = graph->arcs;
  prev = NULL;
  while (curr != NULL && (((arc_t *)curr->data) != arc))
    {
      prev = curr;
      curr = curr->next;
    }

  if (curr == NULL)
    {
      fprintf(stderr, "Arc not found in list!\n");
      abort();
    }

  /* If it is the first element in the list ... */
  if (prev == NULL)
    {
      graph->arcs = curr->next;
    }
  else
    {
      prev->next = curr->next;
    }

  xfree (curr);

  arc_delete(arc);
} /* graph_arc_remove */

void
graph_print(graph_t * graph)
{
  list_t * curr;

  assert(graph != NULL);

  return;
  printf("Arcs:\n");
  for (curr = graph->arcs ; curr != NULL ; curr = curr->next)
    {
      arc_t * arc = (arc_t *) curr->data;
      printf("(%s, %s) = %ld\n", (char *)arc->n1->data,
	     (char *)arc->n2->data, arc->w);
    }
  printf("Nodes:\n");
  for ( curr = graph->nodes ; curr != NULL ; curr = curr->next)
    {
      node_t * node = (node_t *) curr->data;
      printf("%s\n", (char *) node->data);
    }
      

} /* graph_print */

/* Remove a node and all associated arcs from a graph */
int
graph_node_remove(graph_t * graph, node_t * node)
{
  list_t * curr;

  assert(graph != NULL);
  assert(node != NULL);

  if (!list_element_present_p(graph->nodes, node))
    return FALSE;

  /* Remove the node from the list of nodes */
  graph->nodes = list_remove_element(graph->nodes, node);

  /* Remove every arc involving this node */
  for (curr = node->arcs; curr != NULL ;  )
    {
      list_t * next;
      arc_t * arc = (arc_t *) curr->data;

      next = curr->next;

      graph_arc_delete(graph, arc);

      curr = next;
    }

  return TRUE;
} /* graph_node_remove */



