#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "graph.h"

#define WINDOW_SIZE 10

#define xfree(x) { free(x); x = NULL; }

static void
print_node(node_t * node)
{
  printf("node: %s\n", (char *) node->data);
}

static void
add_arcs(graph_t * graph, node_t * dest_node, node_t * source_node,
	 node_t * node)
{
  list_t * curr;

  for (curr = source_node->arcs ; curr != NULL ; curr = curr->next)
    {
      arc_t * curr_arc;
      node_t * other;
      weight_t w;

      curr_arc = (arc_t *) curr->data;

      /* Find the 'other' node */
      if (curr_arc->n1 != source_node)
	other = curr_arc->n1;
      else
	other = curr_arc->n2;

      w = curr_arc->w;

      if (other != node)
	{
	  arc_t * n2_arc;
	  /* Now, find out if 'node' also had an arc with 'other,' and
	     if so, add the weights. */
	  n2_arc = graph_arc_get(graph, other, node);
	  if (n2_arc != NULL)
	    {
	      w += n2_arc->w;
	      graph_arc_delete (graph, n2_arc);
	    }

	  /* Now create a new arc for this new node */
	  graph_arc_new(graph, dest_node, other, w, NULL);
	}
    }

} /* add_arcs */

static void
process(graph_t * graph)
{
  arc_t * arc;
  node_t * new_node;
  char * name;

  /* First print the graph */
  graph_print(graph);

  printf("\n\n\n");
  /* While there are still arcs... */
  while (graph->arcs != NULL)
    {
      node_t * node1, * node2;
      /* Grab the highest arc in the graph */
      graph_arc_sort(graph);
      arc = (arc_t *) graph->arcs->data;

      /* Create a new node which is the combination of the two nodes
	 which made up the highest arc */
      name = malloc(strlen((char *)arc->n1->data) +
		    strlen((char *)arc->n2->data) +
		    2);
      strcpy(name, (char *)arc->n1->data);
      strcat(name, ",");
      strcat(name, (char *)arc->n2->data);

      new_node = node_new(0, name);
      graph_node_add(graph, new_node);

      /* Add arcs for this new node.  This means that, for each of
	 the two nodes that made up the highest arc, we examine every
	 arc which connected to that node. FIXME */
      add_arcs(graph, new_node, arc->n1, arc->n2);
      add_arcs(graph, new_node, arc->n2, arc->n1);

      /* Now delete the old nodes */
      node1 = arc->n1;
      node2 = arc->n2;
      graph_arc_delete(graph, arc);
      graph_node_remove(graph, node1);
      graph_node_remove(graph, node2);
      xfree(node1);
      xfree(node2);

      printf("Completed a grouping...\n");
    }

  graph_print(graph);
  
} /* process */

/* return the node corresponding to the specified string */
static node_t *
string2node(graph_t * graph, char * str)
{
  list_t * curr, * nodes;
  node_t * node;

  assert(str != NULL);

  nodes = graph->nodes;

  curr = nodes;
  while (curr != NULL)
    {
      node = (node_t *) curr->data;
      if (!strcmp((char *)node->data, str))
	return node;

      curr = curr->next;
    }

  node = node_new(0, strdup(str));

  graph_node_add(graph, node);
  return node;
} /* string2node */

static void
window_add_node(list_t ** window, int * window_size, node_t * node)
{
  list_t * new;

  /* Add new nodes to the front of the list, so remove nodes from the
     back of the list */
  if (*window_size >= WINDOW_SIZE)
    {
      list_t * curr, * prev;



      /* Remove a node from the back of the list */
      curr = *window;
      prev = NULL;
      while (curr->next != NULL)
	{
	  prev = curr;
	  curr = curr->next;
	}
      xfree(curr);
      prev->next = NULL;

      (*window_size) --;
    }

  /* Add this node to the beginning of the list */
  new = list_new(node);
  new->next = *window;
  *window = new;
  (*window_size) ++;
} /* window_add_node */

static void
window_print(list_t * window, int window_size)
{
  list_t * curr;
  node_t * node;
  int i;

  /*  printf("Window Size: %d\n", window_size); */
  i = 0;
  for (curr = window; curr != NULL ; curr = curr->next)
    {
      node = (node_t *)curr->data;
      /*      printf("Window[%d]: %s\n", i, (char *)node->data); */
      i++;
    }
  
} /* window_print */

int
main(int argc, char ** argv)
{
  char line[1024];
  graph_t * graph = NULL;
  long lines_processed = 0;

  /* This list will contain the last 'window_size' nodes */
  list_t * window = NULL;

  int window_size = 0;

  assert(WINDOW_SIZE > 2);

  graph = graph_new();

  while (!feof(stdin))
    {
      list_t * curr1, * curr2;
      char * p;

      if (fgets(line, sizeof(line), stdin) == NULL)
	{
	  process(graph);
	  return 0;
	}

      lines_processed ++;
      if ((lines_processed % 1000) == 0)
	fprintf(stderr, "%ld...\n", lines_processed);

      /* Remove the trailing newline */
      p = strchr(line, '\n');
      if (p != NULL)
	*p = '\0';

      window_add_node(&window, &window_size, string2node(graph, line));

      window_print(window, window_size);

      if (window_size == WINDOW_SIZE) 
	{
	  /* Next, make arcs between all nodes in the window */
	  for (curr1 = window ; curr1 != NULL ; curr1 = curr1->next)
	    for (curr2 = window ; curr2 != NULL ; curr2 = curr2->next)
	      {
		arc_t * arc;
		if (curr1->data != curr2->data)
		  {
#if 0
		    printf("arc \"%s\" <==> \"%s\"\n",
			   (char *) ((node_t *)curr1->data)->data,
			   (char *) ((node_t *)curr2->data)->data);
#endif
		    graph_arc_inc(graph, (node_t *)curr1->data,
				  (node_t *)curr2->data,
				  1);
		    arc = graph_arc_get(graph, (node_t *)curr1->data,
					(node_t *)curr2->data);
		  }
	      }
	}
    }

  process(graph);

  return 0;
} /* main */
