/*
 * COPYRIGHT NOTICE
 * 
 * Copyright (c) 1990, 1991, 1992, 1993 Open Software Foundation, Inc. 
 * 
 * Permission is hereby granted to use, copy, modify and freely distribute
 * the software in this file and its documentation for any purpose without
 * fee, provided that the above copyright notice appears in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation.  Further, provided that the name of Open
 * Software Foundation, Inc. ("OSF") not be used in advertising or
 * publicity pertaining to distribution of the software without prior
 * written permission from OSF.  OSF makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */
/*
 * ELF object file dumper.
 */
/*
 * HISTORY
 * $Log: elfdump.c,v $
 * 
 * Nonrevisiones 1996/03  miguel@nuclecu.unam.mx
 * added dumping facilities for SPARC elf executables.
 * 
 * Revision 1.23  1995/03/01  19:49:33  meissner
 * Correctly deal with bi-endian data in print_section_hash.
 *
 * Revision 1.22  1995/02/24  16:43:36  meissner
 * Don't compare unsigneds with signeds.
 * Eliminate STT_STABS hack from OSF/1.
 *
 * Revision 1.21  1995/02/24  14:52:05  meissner
 * Fall back to use V.4's dis if objdump isn't available.
 *
 * Revision 1.20  1995/02/14  21:10:42  meissner
 * Do endian transformation on stab string offsets.
 * Before writing to stderr, fflush stdout.
 *
 * Revision 1.19  1995/02/14  18:56:01  meissner
 * EM_PPC is now 20, not 17.  Reconize both old and new values.
 *
 * Revision 1.18  1995/01/28  02:35:01  meissner
 * Fix spacing if raw section size is not a multiple of 4.
 * For -d, also look for ../../binutils/objdump.
 *
 * Revision 1.17  1995/01/27  17:14:15  meissner
 * Add new PowerPC section relative relocations.
 *
 * Revision 1.16  1995/01/23  16:21:38  meissner
 * Make -e deal with mixed endian systems.
 * Add support for new linker symbols.
 *
 * Revision 1.15  1995/01/22  01:52:44  meissner
 * When writing the layout, size any holes as well.
 *
 * Revision 1.14  1995/01/20  17:19:18  meissner
 * For -d, try using ../binutils/objdump first in case we are in the build tree.
 *
 * Revision 1.13  1995/01/19  22:31:10  meissner
 * Add PowerPC relocations.
 *
 * Revision 1.12  1995/01/19  18:02:32  meissner
 * Replace uchar with Uchar.
 *
 * Revision 1.11  1995/01/19  17:50:22  meissner
 * Allow ELF file to be unaligned.
 * Deal with converting numbers from host format to target format.
 * Remove OSF/1 profile support.
 *
 * Revision 1.10  1994/12/29  17:39:07  meissner
 * Add support to print PowerPC, MIPS, and PA RISC specific flags.
 *
 * Revision 1.9  1994/12/29 16:35:43  meissner
 * Use fflush(stdout); fflush(stderr); instead of fflush(NULL).
 * Don't invoke /usr/ccs/gcc-elf/objdump.
 *
 * Revision 1.8  1994/12/27 19:24:53  meissner
 * Make version string const.
 *
 * Revision 1.7  1994/12/27 18:40:30  meissner
 * Remove support for EM_CYGNUS_POWERPC.
 * Always print value of ehdr fields, not just symbolic name.
 * Print file modification time.
 *
 * Revision 1.6  1994/12/23 18:41:57  meissner
 * Add EM_PPC (powerpc) support.
 *
 * Revision 1.5  1994/12/21 17:47:32  meissner
 * Compile cleanly with -Wall on unixware.
 *
 * Revision 1.4  1994/12/21 17:44:27  meissner
 * Make -Wall silent on Suns.
 *
 * Revision 1.3  1994/12/21 16:32:32  meissner
 * Add elf.h declarations from System V ABI.
 * Make work on SPARC.
 * If NO_MMAP is defined, don't use mmap.
 * Add cygnus local declarations.
 *
 * Revision 1.2  1994/12/20 21:14:03  meissner
 * Move OSF changelog to separate comment.
 *
 * Revision 1.1  1994/12/20 21:11:44  meissner
 * Initial version.
 */
/*
 * Changes made at Open Software Foundation.
 *
 * Revision 1.1.2.32  1994/04/08  17:10:34  meissner
 * 	Correctly print out .lprofil sections when a double word count is used.
 * 	[1994/04/08  17:10:25  meissner]
 *
 * Revision 1.1.2.31  1994/04/07  13:56:54  meissner
 * 	Range check section section index in get_section_name.
 * 	Add program header regions to vm/file map printout.
 * 	[1994/04/07  13:56:44  meissner]
 * 
 * Revision 1.1.2.30  1994/03/17  13:28:57  meissner
 * 	Cut down on more spaces in output.
 * 	Print section name instead of number in symbol.
 * 	If -e, add linker generated symbols in vm layout.
 * 	Indicate where there are holes in the file/vm layout.
 * 	Don't print profil buckets that are 0.
 * 	[1994/03/17  13:28:49  meissner]
 * 
 * Revision 1.1.2.29  1994/03/16  22:27:11  meissner
 * 	Print out layout information for the ELF file and virtual memory.
 * 	Check layouts printed for overlaps.
 * 	Print out end offsets in the program headers.
 * 	Lowercase section flags.
 * 	Cut down on symbol line length.
 * 	[1994/03/16  22:27:02  meissner]
 * 
 * Revision 1.1.2.28  1994/03/16  18:25:24  meissner
 * 	Align summary fields of section headers.
 * 	Print out end+1 offset in section headers.
 * 	Always print all section header fields.
 * 	Skip profiling sections if -p.
 * 	[1994/03/16  18:25:14  meissner]
 * 
 * Revision 1.1.2.27  1994/03/11  22:35:01  chasb
 * 	Expand Copyrights
 * 	[1994/03/09  19:43:03  chasb]
 * 
 * Revision 1.1.2.26  1994/02/28  21:36:34  meissner
 * 	Correctly print out profil lowpc.
 * 	[1994/02/28  21:36:24  meissner]
 * 
 * Revision 1.1.2.25  1994/02/25  16:55:19  meissner
 * 	Add support for printing profile stats on OSF/1.
 * 	Allow unsigned long in addition to int for profil sizes.
 * 	[1994/02/25  16:55:11  meissner]
 * 
 * Revision 1.1.2.24  1993/12/23  01:18:10  meissner
 * 	On OSF/1 print the uarea for profiling ELF files.
 * 	[1993/12/23  01:18:03  meissner]
 * 
 * Revision 1.1.2.23  1993/11/22  21:14:23  meissner
 * 	Add -v option to print the version string.
 * 	[1993/11/22  21:14:16  meissner]
 * 
 * Revision 1.1.2.22  1993/11/22  21:00:50  meissner
 * 	Don't print section name for reloc to defined symbol, print symbol name.
 * 	[1993/11/22  21:00:42  meissner]
 * 
 * Revision 1.1.2.21  1993/11/11  21:40:27  meissner
 * 	Correctly align gprof/prof output.
 * 	Don't print profil ticks in hex.
 * 	[1993/11/11  21:40:12  meissner]
 * 
 * Revision 1.1.2.20  1993/11/11  21:29:19  meissner
 * 	CR 77 - Add profile output support.
 * 	[1993/11/11  21:29:06  meissner]
 * 
 * Revision 1.1.2.19  1993/10/15  18:55:56  meissner
 * 	Fix mmap call to set the MAP_PRIVATE flag.
 * 	[1993/10/15  18:55:44  meissner]
 * 
 * Revision 1.1.2.18  1993/09/23  21:52:22  meissner
 * 	Treat .debug_sfnames as a string section.
 * 	[1993/09/23  21:52:09  meissner]
 * 
 * Revision 1.1.2.17  1993/09/23  21:29:12  meissner
 * 	Skip printing null bytes in string sections, except for the first byte.
 * 	[1993/09/23  21:28:58  meissner]
 * 
 * Revision 1.1.2.16  1993/09/23  21:21:20  meissner
 * 	Print .stabstr section as strings.
 * 	Know about .stab section special format.
 * 	[1993/09/23  21:21:07  meissner]
 * 
 * Revision 1.1.2.15  1993/09/08  22:32:47  meissner
 * 	CR 62: Add -t {d,o,x} to specify output formats.
 * 	[1993/09/08  22:32:35  meissner]
 * 
 * Revision 1.1.2.14  1993/07/27  20:37:07  meissner
 * 	Print 386 relocations in symbolic format.
 * 	[1993/07/27  20:36:52  meissner]
 * 
 * Revision 1.1.2.13  1993/07/23  15:07:30  meissner
 * 	Add definition of STT_STABS if not provided in elf.h.
 * 	Remove ifdefs of STT_STABS elsewhere.
 * 	If -d, invoke objdump -d -j <section> to disassemble executable sections.
 * 	[1993/07/23  15:07:12  meissner]
 * 
 * Revision 1.1.2.12  1993/07/21  01:33:23  meissner
 * 	Print stabs type symbolically, not numerically.
 * 	Rename ELF32_STABS_UNUSED -> ELF32_STABS_OTHER.
 * 	[1993/07/21  01:33:05  meissner]
 * 
 * 	Left align size field, instead of right align.
 * 	[1993/07/18  17:18:12  meissner]
 * 
 * 	Add appropriate casts and/or 'l' specifiers to *printf formats.
 * 	Add STABS in symbol support.
 * 	[1993/07/18  15:07:36  meissner]
 * 
 * Revision 1.1.2.11  1993/06/14  02:15:26  meissner
 * 	Print hash tables again after last change.
 * 	Print interp section in string form, even if -p.
 * 	[1993/06/14  02:15:12  meissner]
 * 
 * Revision 1.1.2.10  1993/06/14  02:02:20  meissner
 * 	CR 52:
 * 	Add -n <name> switch to print a specific section.
 * 	Add -p switch to not print the contents of a progbits section.
 * 	Add -s switch to print header summary only.
 * 	Add -x switch to assume the string table is corrupt.
 * 	[1993/06/14  02:02:04  meissner]
 * 
 * Revision 1.1.2.9  1993/05/27  18:56:26  meissner
 * 	Remove 386 hack which was caused by a buggy compiler.
 * 	[1993/05/27  18:56:11  meissner]
 * 
 * Revision 1.1.2.8  1993/05/24  11:46:12  meissner
 * 	Hack around i386 problem with case SHN_ABS/SHN_COMMON.
 * 	[1993/05/24  11:45:53  meissner]
 * 
 * Revision 1.1.2.7  1993/05/13  17:38:41  meissner
 * 	Add DT_RELAENT.
 * 	[1993/05/13  17:38:24  meissner]
 * 
 * Revision 1.1.2.6  1993/05/12  19:40:17  meissner
 * 	Fix printing of implicit addend in relocation to get full word, not byte.
 * 	Print section being referenced in relocations, symbol tables and hash tables.
 * 	Print name of relocated symbol or section.
 * 	Don't assume REL implicit addend is word aligned.
 * 	Print the dynamic section in a friendly fashion.
 * 	[1993/05/12  19:39:57  meissner]
 * 
 * Revision 1.1.2.5  1993/05/10  21:01:48  meissner
 * 	Fixup so it compiles cleanly on V.4.
 * 	Don't abort if implicit addend too large.
 * 	Fix GCC warnings.
 * 	Relocations for executables have virtual address in r_offset, not section offset.
 * 	Split sym/type in relocations.
 * 	[1993/05/10  21:01:28  meissner]
 * 
 * Revision 1.1.2.4  1993/05/10  15:21:18  meissner
 * 	Print REL implicit addend.
 * 	[1993/05/10  15:20:59  meissner]
 * 
 * Revision 1.1.2.3  1993/05/10  13:38:56  meissner
 * 	Change copyright to OSF_FREE_COPYRIGHT.
 * 	Deal with MAP_VARIABLE not being defined.
 * 	Add hash table support.
 * 	[1993/05/10  13:38:32  meissner]
 * 
 * Revision 1.1.2.2  1993/05/07  19:44:44  meissner
 * 	Initial version.
 * 	[1993/05/07  19:44:20  meissner]
 */

static const char elfdump_what_string[] =
  "@(#)elfdump.c\t$Revision: 1.23 $ $Date: 1995/03/01 19:49:33 $ $Locker:  $";

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>

#ifndef NO_MMAP
#include <sys/mman.h>

#if !defined(MAP_PRIVATE) || !defined(MAP_FILE) || !defined(MAP_VARIABLE)
#define	NO_MMAP
#endif
#endif

#ifdef __OSF1__
#include <sys/time.h>
#include <sys/user.h>
#endif

#ifdef __sparc__
extern void perror (const char *);
extern int printf (const char *, ...);
extern int fprintf (FILE *, const char *, ...);
extern int fflush (FILE *);
extern int tolower (int);
extern int _flsbuf (int, FILE *);
#endif

extern int open (const char *, int, ...);

typedef unsigned char Uchar;

#ifndef DIGIT_OUTPUT_SIZE
#define DIGIT_OUTPUT_SIZE	64	/* buffer size needed to format integers, includes prefix/suffix */
#endif

#ifndef	MAX_FLAGS_SIZE
#define	MAX_FLAGS_SIZE		1024	/* buffer size for printing the flags */
#endif

extern time_t time (time_t *);
extern int getopt (int, char * const [], const char *);
extern char *optarg;
extern int optind;
extern int optopt;
extern int opterr;


/*
 * Data definitions from the System V ABI.
 */

/* 32-bit data types */
#if ULONG_MAX > 0xFFFFFFFF			/* 64 bit systems */
typedef unsigned int	Elf32_Addr;
typedef unsigned short	Elf32_Half;
typedef unsigned int	Elf32_Off;
typedef int		Elf32_Sword;
typedef unsigned int	Elf32_Word;

#else						/* 32 bit systems */
typedef unsigned long	Elf32_Addr;
typedef unsigned short	Elf32_Half;
typedef unsigned long	Elf32_Off;
typedef long		Elf32_Sword;
typedef unsigned long	Elf32_Word;
#endif

/* e_ident[] Identification Indexes, pages 4-7 .. 4-9 */
#define	EI_MAG0		0			/* magic number, byte 0 */
#define	EI_MAG1		1			/* magic number, byte 1 */
#define	EI_MAG2		2			/* magic number, byte 2 */
#define	EI_MAG3		3			/* magic number, byte 3 */
#define	EI_CLASS	4			/* 32-bit or 64-bit */
#define	EI_DATA		5			/* data encoding */
#define	EI_VERSION	6			/* file version number */
#define	EI_PAD		7			/* start of padding bytes */
#define	EI_NIDENT	16			/* size of e_ident header */

#define ELFMAG0	       	'\x7f'			/* magic number, byte 0 */
#define	ELFMAG1		'E'			/* magic number, byte 1 */
#define	ELFMAG2		'L'			/* magic number, byte 2 */
#define	ELFMAG3		'F'			/* magic number, byte 3 */

#define	ELFCLASSNONE	0			/* invalid class */
#define	ELFCLASS32	1			/* 32-bit objects */
#define	ELFCLASS64	2			/* 64-bit objects */

#define	ELFDATANONE	0			/* invalid data encoding */
#define	ELFDATA2LSB	1			/* little endian */
#define	ELFDATA2MSB	2			/* big endian */

/* Object file types, page 4-5 */
#define	ET_NONE		0			/* no file type */
#define	ET_REL		1			/* relocatable file */
#define	ET_EXEC		2			/* executable file */
#define	ET_DYN		3			/* shared object file */
#define	ET_CORE		4			/* core file */
#define	ET_LOPROC	0xff00			/* low  processor specific */
#define	ET_HIPROC	0xffff			/* high processor specific */

/* Machine type, page 4-5 */
#define EM_NONE			0		/* no machine */
#define	EM_M32			1		/* AT&T WE 32100 */
#define	EM_SPARC		2		/* SPARC */
#define	EM_386			3		/* Intel 80386 */
#define	EM_68K			4		/* Motorola 68000 */
#define	EM_88K			5		/* Motorola 88000 */
#define	EM_860			7		/* Intel 80860 */
#define	EM_MIPS			8		/* Mips RS3000 */
#define EM_CYGNUS_POWERPC	0x9025		/* Private Cygnus PowerPC */
#define EM_CYGNUS_ARC		0x9026		/* Private Cygnus ARC */

/* Some of these are not offical or are for internal cygnus ports */
#define EM_MIPS_RS4_BE		10		/* MIPS R4000 big-endian */
#define EM_SPARC64		11		/* SPARC v9 (not official) 64-bit */
#define EM_PARISC		15		/* HPPA */
#define	EM_PPC			20		/* PowerPC (svr4 abi) */
#define EM_PPC_OLD		17		/* old version of PowerPC */

/* Machine dependent flags */
#define	EF_PPC_EMB		0x80000000	/* PowerPC embedded flag  */

#define EF_MIPS_NOREORDER	0x00000001	/* At least one .noreorder directive appears in the source.  */
#define EF_MIPS_PIC		0x00000002	/* File contains position independent code.  */
#define EF_MIPS_CPIC		0x00000004	/* Code in file uses the standard calling sequence for calling
						   position independent code.  */

#define EF_MIPS_ARCH		0xf0000000	/* Four bit MIPS architecture field.  */
#define EF_MIPS_ARCH_1		0x00000000	/* -mips1 code.  */
#define EF_MIPS_ARCH_2		0x10000000	/* -mips2 code.  */
#define EF_MIPS_ARCH_3		0x20000000	/* -mips3 code.  */

#define	EF_PARISC_ARCH		0x0000ff7f	/* PA 1.0/1.1 field */
#define	EF_PARISC_ARCH_1_0	0x0000020b	/* PA 1.0 big endian.  */
#define	EF_PARISC_ARCH_1_1	0x00000210	/* PA 1.1 big endian.  */
#define	EF_PARISC_ENDIAN	0x00000080	/* PA endian field */
#define	EF_PARISC_ENDIAN_BIG	0x00000000	/* PA big endian */
#define	EF_PARISC_ENDIAN_LITTLE	0x00000080	/* PA little endian */
#define EF_PARISC_TRAPNIL	0x00010000	/* Trap null address dereferences.  */
#define EF_PARISC_EXT		0x00020000	/* .PARISC.archext section is present.  */


/* Object file version number, page 4-6 */
#define	EV_NONE		0			/* Invalid version */
#define	EV_CURRENT	1			/* Current version */

/* Elf header, page 4-4 */
typedef struct {
	unsigned char	e_ident[EI_NIDENT];	/* describe ELF layout */
	Elf32_Half	e_type;			/* object file type */
	Elf32_Half	e_machine;		/* machine architecture */
	Elf32_Word	e_version;		/* object file version number */
	Elf32_Addr	e_entry;		/* start address */
	Elf32_Off	e_phoff;		/* program header offset */
	Elf32_Off	e_shoff;		/* section table offset */
	Elf32_Word	e_flags;		/* machine dependent flags */
	Elf32_Half	e_ehsize;		/* elf header size */
	Elf32_Half	e_phentsize;		/* sizeof(Elf32_Phdr) */
	Elf32_Half	e_phnum;		/* # of program headers */
	Elf32_Half	e_shentsize;		/* sizeof(Elf32_Shdr) */
	Elf32_Half	e_shnum;		/* # of section headers */
	Elf32_Half	e_shstrndx;		/* section name table index */
} Elf32_Ehdr;

/* Section indexes, page 4-11 */
#define	SHN_UNDEF	0			/* no section */
#define	SHN_LORESERVE	0xff00			/* low bound of reserved indexes */
#define	SHN_LOPROC	0xff00			/* low bound of processor reserved */
#define	SHN_HIPROC	0xff1f			/* high bound of processor reserved */
#define	SHN_ABS		0xfff1			/* absolute value */
#define	SHN_COMMON	0xfff2			/* common symbol */
#define	SHN_HIRESERVE	0xffff			/* high bound of reserved indexes */

/* Section types, page 4-15 */
#define	SHT_NULL	0			/* inactive section */
#define	SHT_PROGBITS	1			/* program bits */
#define	SHT_SYMTAB	2			/* symbol table */
#define	SHT_STRTAB	3			/* string table */
#define	SHT_RELA	4			/* RELA relocations */
#define	SHT_HASH	5			/* hash table */
#define	SHT_DYNAMIC	6			/* dynamic linking section */
#define	SHT_NOTE	7			/* notes */
#define	SHT_NOBITS	8			/* BSS section */
#define	SHT_REL		9			/* REL relocations */
#define	SHT_SHLIB	10			/* reserved for shared libraries */
#define	SHT_DYNSYM	11			/* dynamic symbol table */
#define	SHT_LOPROC	0x70000000		/* low  bounds, processor specific */
#define	SHT_HIPROC	0x7fffffff		/* high bounds, processor specific */
#define	SHT_LOUSER	0x80000000		/* low  bounds, user reserved */
#define	SHT_HIUSER	0xffffffff		/* high bounds, user reserved */

/* Section attribute flags, page 4-18 */
#define	SHF_WRITE	0x1			/* section is writable */
#define	SHF_ALLOC	0x2			/* section is allocated in memory */
#define	SHF_EXECINSTR	0x4			/* section contains instructions */
#define	SHF_MASKPROC	0xf0000000		/* mask for processor specific flags */

/* Section Header, page 4-13 */
typedef struct {
	Elf32_Word	sh_name;		/* section name */
	Elf32_Word	sh_type;		/* section type */
	Elf32_Word	sh_flags;		/* section flags */
	Elf32_Addr	sh_addr;		/* address section is loaded at or 0 */
	Elf32_Off	sh_offset;		/* byte offset of section */
	Elf32_Word	sh_size;		/* section size */
	Elf32_Word	sh_link;		/* section table index */
	Elf32_Word	sh_info;		/* extra information */
	Elf32_Word	sh_addralign;		/* section alignment (power of 2) */
	Elf32_Word	sh_entsize;		/* size of fixed fields in section */
} Elf32_Shdr;


/* Symbol binding, page 4-26 */
#define	STB_LOCAL	0			/* local symbol */
#define	STB_GLOBAL	1			/* global symbol */
#define	STB_WEAK	2			/* weak symbol */
#define	STB_LOPROC	13			/* low bound, processor reserved */
#define	STB_HIPROC	15			/* high bound, processor reserved */

/* Symbol type, page 4-28 */
#define	STT_NOTYPE	0			/* no type information supplied */
#define	STT_OBJECT	1			/* symbol is an object */
#define	STT_FUNC	2			/* symbol is a function */
#define	STT_SECTION	3			/* symbol is a section */
#define	STT_FILE	4			/* symbol is a filename */
#define	STT_LOPROC	13			/* low bound, processor reserved */
#define	STT_HIPROC	15			/* high bound, processor reserved */

/* Macros to encode/decode stabs in ELF, OSF/1 extension */
#define ELF32_STABS_OTHER(s)	((s) >> 24)
#define ELF32_STABS_TYPE(s)	(((s) >> 16) & 0xff)
#define ELF32_STABS_DESC(s)	((short) s)
#define ELF32_STABS_SIZE(u,t,d)	(((u) << 24) | (((t) & 0xff) << 16) | ((d) & 0xffff))

/* Macros to encode/decode binding and type, page 4-26 */
#define	ELF32_ST_BIND(i)	((i) >> 4)		/* decode binding info */
#define	ELF32_ST_TYPE(i)	((i) & 0xf)		/* decode type info */
#define	ELF32_ST_INFO(b,t)	(((b)<<4) + ((t)&0xf))	/* encode binding, type */

/* Symbol table, page 4-25 */
typedef struct {
	Elf32_Word	st_name;		/* symbol name */
	Elf32_Addr	st_value;		/* symbol value */
	Elf32_Word	st_size;		/* symbol size */
	Uchar		st_info;		/* symbol's type & binding */
	Uchar		st_other;		/* reserved */
	Elf32_Half	st_shndx;		/* symbol's section */
} Elf32_Sym;

/* Macros to encode/decode relocation type and symbol, page 4-32 */
#define	ELF32_R_SYM(i)		((i)>>8)		/* decode symbol number */
#define	ELF32_R_TYPE(i)		((Uchar)(i))		/* decode relocation type */
#define	ELF32_R_INFO(s,t)	(((s)<<8) + (Uchar)(t))	/* encode */

/* Relocation entries, page 4-31 */
typedef struct {				/* REL relocations, addend in file */
	Elf32_Addr	r_offset;		/* location to apply relocation */
	Elf32_Word	r_info;			/* relocation type + symbol index */
} Elf32_Rel;

typedef struct {				/* RELA relocations, addend here */
	Elf32_Addr	r_offset;		/* location to apply relocation */
	Elf32_Word	r_info;			/* relocation type + symbol index */
	Elf32_Sword	r_addend;		/* value to add to symbol */
} Elf32_Rela;

/* Program header types, page 5-3 */
#define	PT_NULL		0			/* unused program header */
#define	PT_LOAD		1			/* loadable segment */
#define	PT_DYNAMIC	2			/* dynamic link information */
#define	PT_INTERP	3			/* interpreter section */
#define	PT_NOTE		4			/* notes */
#define	PT_SHLIB	5			/* reserved for shared libraries */
#define	PT_PHDR		6			/* program header itself */
#define	PT_LOPROC	0x70000000		/* low bound, processor reserved */
#define	PT_HIPROC	0x7fffffff		/* high bound, processor reserved */

/* Segment permissions, page 5-6 */
#define	PF_X		0x1			/* execute */
#define	PF_W		0x2			/* write */
#define	PF_R		0x4			/* read */
#define	PF_MASKPROC	0xf0000000		/* processor reserved */

/* Program header, page 5-2 */
typedef struct {
	Elf32_Word	p_type;			/* which program header this is */
	Elf32_Off	p_offset;		/* segment offset */
	Elf32_Addr	p_vaddr;		/* virtual address */
	Elf32_Addr	p_paddr;		/* physical address */
	Elf32_Word	p_filesz;		/* size of file segment */
	Elf32_Word	p_memsz;		/* size of memory segment */
	Elf32_Word	p_flags;		/* flags */
	Elf32_Word	p_align;		/* alignment */
} Elf32_Phdr;

/* dynamic structure - page 5-15, figure 5-9 */
typedef struct {
	Elf32_Sword	d_tag;
	union {
	    Elf32_Word	d_val;
	    Elf32_Addr	d_ptr;
	} d_un;
} Elf32_Dyn;

/* Dynamic array tags - page 5-16, figure 5-10.  */
#define DT_NULL		0
#define DT_NEEDED	1
#define DT_PLTRELSZ	2
#define DT_PLTGOT	3
#define DT_HASH		4
#define DT_STRTAB	5
#define DT_SYMTAB	6
#define DT_RELA		7
#define DT_RELASZ	8
#define DT_RELAENT      9
#define DT_STRSZ	10
#define DT_SYMENT	11
#define DT_INIT		12
#define DT_FINI		13
#define DT_SONAME	14
#define DT_RPATH	15
#define DT_SYMBOLIC	16
#define DT_REL		17
#define DT_RELSZ	18
#define DT_RELENT	19
#define DT_PLTREL	20
#define DT_DEBUG	21
#define DT_TEXTREL	22
#define DT_JMPREL	23


struct sec_list {
  struct sec_list	*next;
  const char		*section;
};

typedef struct elf_stab {
  Elf32_Off		soffset;			/* offset to string in .stabstr */
  unsigned char		type;				/* stab type */
  unsigned char		other;				/* stab other field */
  Elf32_Half		desc;				/* stab desc field */
  Elf32_Sword		value;				/* stab value field */
} elf_stab_t;

static char		*file;				/* current file being processed */
static Uchar		*file_bytes = (Uchar *)-1;	/* pointer to current object file data */
static struct stat	 file_stat;			/* file status information */
static int		 file_fd	= -1;		/* file descriptor file is openned on */
static Elf32_Ehdr	 ehdr;				/* global header */
static Elf32_Shdr	*fixed_sect_hdrs;		/* pointer to fixed section headers */
static Elf32_Phdr	*fixed_prog_hdrs;		/* pointer to fixed section headers */
static struct sec_list	*sections;			/* sections to print out */
static int		 ignore_string	= 0;		/* ignore the string table */
static int		 objdump	= 0;		/* invoke objdump -d to disassemble text */
static int		 no_progbits	= 0;		/* skip printing progbits section */
static int		 no_layout	= 0;		/* skip printing file/vm layout information */
static int		 summary	= 0;		/* print just summary */
static int		 linker_symbols = 0;		/* put etext/edata/end into layout */
static int		 print_vers	= 0;		/* print version string */
static int		 print_header	= 0;		/* print just raw elf header */
static int		 alloc_file	= 0;		/* file_bytes was allocated */
static int		 errors		= 0;		/* # errors found */
static char		 format		= '\0';		/* default format to use */
static int		 sh_name_width	= 0;		/* width of section names */
static int		 sh_type_width  = 0;		/* width of section types */
static int		 sh_start_width	= 0;		/* width of section offset start */
static int		 sh_end_width	= 0;		/* width of section offset end */
static int		 sh_size_width	= 0;		/* width of section size */
static int		 sh_addr_width	= 0;		/* width of section address */
static int		 sh_align_width = 0;		/* width of section alignment */
static int		 sh_flags_width = 0;		/* width of section flag bits */
static int		 sh_index_width = 0;		/* width of section index */
static int		 sh_info_width	= 0;		/* width of section info field */
static int		 sh_link_width	= 0;		/* width of section link field */
static int		 sh_esize_width = 0;		/* width of section entsize field */

static	char		 flags_print[MAX_FLAGS_SIZE];	/* printable flags */

#define SEC_PTR(index)	   (&fixed_sect_hdrs[index])
#define PROGHDR_PTR(index) (&fixed_prog_hdrs[index])

static const char *const dyn_tag[] = {
  "null",
  "needed",
  "pltrelsz",
  "pltgot",
  "hash",
  "strtab",
  "symtab",
  "rela",
  "relasz",
  "relaent",
  "strsz",
  "syment",
  "init",
  "fini",
  "soname",
  "rpath",
  "symbolic",
  "rel",
  "relsz",
  "relent",
  "pltrel",
  "debug",
  "textrel",
  "jmprel",
};

/* Enum that says to free name in layout chain */
typedef enum layout_free {
  layout_free_no	= 0,	/* do not free name field */
  layout_free_yes	= 1	/* free name field */
} layout_free_t;

/* Structure that records the information about file/virtual layout */
typedef struct layout {
  struct layout *next;		/* next layout element */
  const char *name;		/* element name */
  unsigned long start;		/* start address */
  unsigned long size;		/* size of region */
  layout_free_t free_p;		/* whether to free name or not */
} layout_t;

static layout_t *file_layout	= (layout_t *)0;
static layout_t *vm_layout	= (layout_t *)0;

static const char *get_section_name (Elf32_Word);

/* Header fields that are printed */
enum hdr_fields {
    HDR_TYPE,				/* type (none, rel, exec, dyn, etc.) */
    HDR_MACHINE,			/* machine (sparc, 386, etc.) */
    HDR_CLASS,				/* class (32bit, 64bit) */
    HDR_DATA,				/* data (little endian, big endian) */
    HDR_VERSION,			/* version # */
    HDR_FLAGS,				/* flags */
    HDR_ENTRY,				/* entry address */
    HDR_EHSIZE,				/* header size */
    HDR_SHSTRNDX,			/* string index */
    HDR_MAX				/* max # headers printed */
};

/* Header fields to be printed */
static struct {
    const char	  *title;		/* title to print */
    const char	  *print;		/* printable version of field */
    unsigned long  value;		/* value to print */
    char	   format;		/* format to use */
} hdr_field[ (int)HDR_MAX ];


/* Lookup table for machine dependent flags */
static struct {
    const char	*string;			/* string version of flags */
    Elf32_Word	 value;				/* value to compare after AND */
    Elf32_Word	 mask;				/* value to AND to get bits */
    Elf32_Half	 machine;			/* machine these flags are valid for */
} elf32_flags[] = {
  { ", embedded",	EF_PPC_EMB,		EF_PPC_EMB,		EM_PPC },
	
  { ", noreorder",	EF_MIPS_NOREORDER,	EF_MIPS_NOREORDER,	EM_MIPS },
  { ", pic",		EF_MIPS_PIC,		EF_MIPS_PIC,		EM_MIPS },
  { ", cpic",		EF_MIPS_CPIC,		EF_MIPS_CPIC,		EM_MIPS },
  { ", mips1",		EF_MIPS_ARCH_1,		EF_MIPS_ARCH,		EM_MIPS },
  { ", mips2",		EF_MIPS_ARCH_2,		EF_MIPS_ARCH,		EM_MIPS },
  { ", mips3",		EF_MIPS_ARCH_3,		EF_MIPS_ARCH,		EM_MIPS },
	
  { ", PA RISC 1.0",	EF_PARISC_ARCH_1_0,	EF_PARISC_ARCH,		EM_PARISC },
  { ", PA RISC 1.1",	EF_PARISC_ARCH_1_1,	EF_PARISC_ARCH,		EM_PARISC },
  { ", big endian",	EF_PARISC_ENDIAN_BIG,	EF_PARISC_ENDIAN,	EM_PARISC },
  { ", little endian",	EF_PARISC_ENDIAN_LITTLE,EF_PARISC_ENDIAN,	EM_PARISC },
  { ", trap nil",	EF_PARISC_TRAPNIL,	EF_PARISC_TRAPNIL,	EM_PARISC },
  { ", archext",	EF_PARISC_EXT,		EF_PARISC_EXT,		EM_PARISC },

  { (const char *)0,	0,			0,			0 }
};


/* Allocate memory, abort if no memory is available.  */

static void *
xmalloc (size_t len)
{
  void *p = malloc (len);
  if (!p)
    {
      perror ("malloc");
      exit (1);
    }

  return p;
}

/* Allocate and clear memory, abort if no memory is available.  */

static void *
xcalloc (size_t len1, size_t len2)
{
  void *p = calloc (len1, len2);
  if (!p)
    {
      perror ("calloc");
      exit (1);
    }

  return p;
}


/*
 * Convert Elf32_Words from target to host format.
 */

static Elf32_Word decode_elf32_word (Elf32_Word value)
{
    unsigned char *ptr = (unsigned char *)&value;
    Elf32_Word ret;

    switch (ehdr.e_ident[EI_DATA])
      {
      default:		ret = value;							break;
      case ELFDATA2LSB: ret = (ptr[3] << 24) | (ptr[2] << 16) | (ptr[1] << 8) | ptr[0];	break;
      case ELFDATA2MSB: ret = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | ptr[3];	break;
      }

    return ret;
}

#define decode_elf32_sword(value) ((Elf32_Sword) decode_elf32_word ((Elf32_Word)value))
#define decode_elf32_addr(value)  ((Elf32_Addr)  decode_elf32_word ((Elf32_Word)value))
#define decode_elf32_off(value)   ((Elf32_Off)   decode_elf32_word ((Elf32_Word)value))

/*
 * Convert Elf32_Halfs from target to host format.
 */

static Elf32_Half decode_elf32_half (Elf32_Half value)
{
    unsigned char *ptr = (unsigned char *)&value;
    Elf32_Half ret;

    switch (ehdr.e_ident[EI_DATA])
      {
      default:		ret = value;			break;
      case ELFDATA2LSB: ret = (ptr[1] << 8) | ptr[0];	break;
      case ELFDATA2MSB: ret = (ptr[0] << 8) | ptr[1];	break;
      }

    return ret;
}

/*
 * Convert a target symbol table into a host symbol table.  In case the
 * Elf file is not aligned, copy the target structure into a local one
 * first.
 */

static void
decode_elf32_sym (Elf32_Sym *sym, const Elf32_Sym *target_ptr)
{
  Elf32_Sym target_sym;

  memcpy ((void *)&target_sym, (const void *)target_ptr, sizeof (Elf32_Sym));

  sym->st_name  = decode_elf32_word (target_sym.st_name);	/* symbol name */
  sym->st_value = decode_elf32_addr (target_sym.st_value);	/* symbol value */
  sym->st_size  = decode_elf32_word (target_sym.st_size);	/* symbol size */
  sym->st_info  = target_sym.st_info;				/* symbol's type & binding */
  sym->st_other = target_sym.st_other;				/* reserved */
  sym->st_shndx = decode_elf32_half (target_sym.st_shndx);	/* symbol's section */
  return;
}

/*
 * Convert a target REL relocation into a host relocation.  In case the
 * Elf file is not aligned, copy the target structure into a local one
 * first.
 */

static void
decode_elf32_rel (Elf32_Rel *rel, const Elf32_Rel *target_ptr)
{
  Elf32_Rel target_rel;

  memcpy ((void *)&target_rel, (const void *)target_ptr, sizeof (Elf32_Rel));

  rel->r_offset = decode_elf32_addr (target_rel.r_offset);
  rel->r_info   = decode_elf32_word (target_rel.r_info);
  return;
}

/*
 * Convert a target RELA relocation into a host relocation.  In case the
 * Elf file is not aligned, copy the target structure into a local one
 * first.
 */

static void
decode_elf32_rela (Elf32_Rela *rela, const Elf32_Rela *target_ptr)
{
  Elf32_Rela target_rela;

  memcpy ((void *)&target_rela, (const void *)target_ptr, sizeof (Elf32_Rela));

  rela->r_offset = decode_elf32_addr  (target_rela.r_offset);
  rela->r_info   = decode_elf32_word  (target_rela.r_info);
  rela->r_addend = decode_elf32_sword (target_rela.r_addend);
  return;
}

/*
 * Convert a target section header to a host section header.  In case the
 * Elf file is not aligned, copy the target structure into a local one
 * first.
 */

static void
decode_elf32_shdr (Elf32_Shdr *shdr, const Elf32_Shdr *target_ptr)
{
  Elf32_Shdr target_shdr;

  memcpy ((void *)&target_shdr, (const void *)target_ptr, sizeof (Elf32_Shdr));

  shdr->sh_name      = decode_elf32_word (target_shdr.sh_name);
  shdr->sh_type      = decode_elf32_word (target_shdr.sh_type);
  shdr->sh_flags     = decode_elf32_word (target_shdr.sh_flags);
  shdr->sh_addr      = decode_elf32_addr (target_shdr.sh_addr);
  shdr->sh_offset    = decode_elf32_off  (target_shdr.sh_offset);
  shdr->sh_size      = decode_elf32_word (target_shdr.sh_size);
  shdr->sh_link      = decode_elf32_word (target_shdr.sh_link);
  shdr->sh_info      = decode_elf32_word (target_shdr.sh_info);
  shdr->sh_addralign = decode_elf32_word (target_shdr.sh_addralign);
  shdr->sh_entsize   = decode_elf32_word (target_shdr.sh_entsize);
  return;
}

/*
 * Convert a target program header to a host program header.  In case the
 * Elf file is not aligned, copy the target structure into a local one
 * first.
 */

static void
decode_elf32_phdr (Elf32_Phdr *phdr, const Elf32_Phdr *target_ptr)
{
  Elf32_Phdr target_phdr;

  memcpy ((void *)&target_phdr, (const void *)target_ptr, sizeof (Elf32_Phdr));

  phdr->p_type   = decode_elf32_word (target_phdr.p_type);
  phdr->p_offset = decode_elf32_off  (target_phdr.p_offset);
  phdr->p_vaddr  = decode_elf32_addr (target_phdr.p_vaddr);
  phdr->p_paddr  = decode_elf32_off  (target_phdr.p_paddr);
  phdr->p_filesz = decode_elf32_word (target_phdr.p_filesz);
  phdr->p_memsz  = decode_elf32_word (target_phdr.p_memsz);
  phdr->p_flags  = decode_elf32_word (target_phdr.p_flags);
  phdr->p_align  = decode_elf32_word (target_phdr.p_align);
  return;
}

/*
 * Convert a target dynamic structure to a host dynamic structure.  In case the
 * Elf file is not aligned, copy the target structure into a local one
 * first.
 */

static void
decode_elf32_dyn (Elf32_Dyn *dyn, const Elf32_Dyn *target_ptr)
{
  Elf32_Dyn target_dyn;

  memcpy ((void *)&target_dyn, (const void *)target_ptr, sizeof (Elf32_Dyn));

  dyn->d_tag      = decode_elf32_sword (target_dyn.d_tag);
  dyn->d_un.d_val = decode_elf32_sword (target_dyn.d_un.d_val);
  return;
}


/* Convert a number to a string using a format specified by -t <format> or a default one.  */

#define MAX_STRINGIFY	32	/* if more concurrent calls are made this needs to be bumped */

static const char *
stringify (unsigned long number, int width, int def_format, int skip_lead, const char *prefix, const char *suffix)
{
  static unsigned next_buffer;
  static char def_buffer[MAX_STRINGIFY][DIGIT_OUTPUT_SIZE];

  int use_format = (format) ? format : def_format;
  char *buffer	 = &def_buffer[ (next_buffer++) % MAX_STRINGIFY ][0];

  switch (use_format)
    {
    default:
      sprintf (buffer, "%s%*ld%s", prefix, width, (long) number, suffix);
      break;

    case 'x':
      sprintf (buffer, "%s%s%0*lx%s", prefix, (skip_lead) ? "" : "0x", width, number, suffix);
      break;

    case 'o':
      sprintf (buffer, "%s%s%0*lo%s", prefix, (skip_lead) ? "" : "0", width, number, suffix);
      break;

    case 'u':
      sprintf (buffer, "%s%*lu%s", prefix, width, number, suffix);
      break;
    }

  return buffer;
}

/* Return the length of converting a number to a string via stringify.  */

static int
stringify_len (unsigned long number, int def_format, int skip_lead)
{
  return strlen (stringify (number, 0, def_format, skip_lead, "", ""));
}


/* Silence -Wconversion */
#define stringify(n, w, d, l, p, s) (stringify)((unsigned long)(n), w, d, l, p, s)
#define stringify_len(n, d, l) (stringify_len)((unsigned long)(n), d, l)


/* Add an element to the layout linked list. */

static void
add_layout (layout_t **head_ptr, const char *name, unsigned long start, unsigned long size, layout_free_t free_p)
{
  layout_t *ptr = xmalloc (sizeof (layout_t));

  ptr->next   = *head_ptr;
  ptr->name   = name;
  ptr->start  = start;
  ptr->size   = size;
  ptr->free_p = free_p;
  *head_ptr   = ptr;
}

/* Silence -Wconversion */
#define add_layout(head_ptr, name, start, size, free_p) \
  add_layout (head_ptr, name, (unsigned long)start, (unsigned long)size, free_p)


/*
 * Delete a layout chain
 */

static void
free_layout (layout_t **head_ptr)
{
  layout_t *ptr = *head_ptr;
  layout_t *next;

  *head_ptr = (layout_t *)0;
  while (ptr)
    {
      next = ptr->next;
      if (ptr->free_p)
	free ((void *)ptr->name);

      free ((void *)ptr);
      ptr = next;
    }
}


/* Compare two layout pointers, sorting by start address. */

static int
compare_layout (const void *a, const void *b)
{
  const layout_t *ptr_a = *(const layout_t **)a;
  const layout_t *ptr_b = *(const layout_t **)b;

  return (ptr_a->start != ptr_b->start)
		? ptr_a->start - ptr_b->start
		: ptr_a->size  - ptr_b->size;
}


/* Sort and print a layout linked list */

static void
print_layout (layout_t *head, const char *desc)
{
  size_t num_ptrs;
  size_t i;
  int len;
  layout_t *ptr;
  layout_t **ptr_array;
  int width_start = 0;
  int width_end = 0;
  int width_size = 0;
  unsigned long oend;

  if (!head)
    {
      printf ("\nLayout for %s was empty\n", desc);
      return;
    }

  /* Count the number of pointers */
  num_ptrs = 0;
  for (ptr = head; ptr != (layout_t *)0; ptr = ptr->next)
    num_ptrs++;

  /* Sort the layout pointers */
  ptr_array = (layout_t **) xmalloc (num_ptrs * sizeof (layout_t *));
  i = 0;
  for (ptr = head; ptr != (layout_t *)0; ptr = ptr->next)
    ptr_array[i++] = ptr;

  qsort ((void *)ptr_array, num_ptrs, sizeof (layout_t *), compare_layout);

  /* Size start, end */
  oend = 0;
  for (i = 0; i < num_ptrs; i++)
    {
      ptr = ptr_array[i];
      len = stringify_len (ptr->start, 'x', 1);
      if (len > width_start)
	width_start = len;

      len = stringify_len (ptr->start + ptr->size, 'x', 1);
      if (len > width_end)
	width_end = len;

      len = stringify_len (ptr->size, 'd', 1);
      if (len > width_size)
	width_size = len;

      if (i > 0 && ptr->start > oend)
	{
	  len = stringify_len (ptr->start - oend, 'd', 1);
	  if (len > width_size)
	    width_size = len;
	}

      oend = ptr->start + ptr->size;
    }

  /* Print out the layout now */
  printf ("\nLayout for %s:\n", desc);
  oend = 0;
  for (i = 0; i < num_ptrs; i++)
    {
      ptr = ptr_array[i];
      if (i > 0 && ptr->start > oend)
	printf ("    start= %s, end= %s, size= %s, name= %s",
		stringify (oend, width_start, 'x', 0, "", ""),
		stringify (ptr->start, width_end, 'x', 0, "", ""),
		stringify (ptr->start - oend, width_size, 'd', 0, "", ""),
		"<hole>\n");

      printf ("    start= %s, end= %s, size= %s, name= %s",
	      stringify (ptr->start, width_start, 'x', 0, "", ""),
	      stringify (ptr->start + ptr->size, width_end, 'x', 0, "", ""),
	      stringify (ptr->size, width_size, 'd', 0, "", ""),
	      ptr->name);

      if (i > 0 && ptr->start < oend)
	printf ("%*s**** Overlap!!", (int)strlen (ptr->name) - 32, "");

      putchar ('\n');
      oend = ptr->start + ptr->size;
    }

  free (ptr_array);
}


/* Convert an elf string reference into a real string.  */

static const char *
elf_string (Elf32_Word sec_index, Elf32_Word byte_off)
{
  static char bad_string[] = "???";

  Elf32_Shdr *sec_hdr = SEC_PTR (sec_index);

  if (ignore_string)
    return bad_string;

  if (sec_index <= 0
      || sec_index > (Elf32_Word)ehdr.e_shnum
      || file_bytes == (Uchar *)-1
      || (sec_hdr->sh_type != SHT_STRTAB
	  && strcmp (get_section_name (sec_index), ".stabstr") != 0))
    {
      fflush (stdout);
      fprintf (stderr, "%s section %s is not a string table.\n", file,
	       stringify (sec_index, 0, 'd', 0, "", ""));

      fflush (stderr);
      return bad_string;
    }

  if (byte_off > sec_hdr->sh_size)
    {
      fflush (stdout);
      fprintf (stderr, "%s string offset %s too big for section %s\n", file,
	       stringify (byte_off, 0, 'd', 0, "", ""),
	       stringify (sec_index, 0, 'd', 0, "", ""));

      fflush (stderr);
      return bad_string;
    }

  return (char *)(file_bytes + sec_hdr->sh_offset + byte_off);
}


/* Convert 1..EI_NIDENT bytes into printable notation.  */

static char *
printable_string (Uchar *ptr, int num)
{
  static char print_buffer[(EI_NIDENT * 4) + 4];
  int i;
  int ch;
  char *buffer_ptr = print_buffer;

  *buffer_ptr++ = '"';
  for (i = 0; i < num; i++)
    {
      ch = *ptr++;
      switch (ch)
	{
	default:
	  if (isprint (ch))
	    *buffer_ptr++ = ch;

	  else
	    {
	      sprintf (buffer_ptr, "\\%03o", ch);
	      buffer_ptr += 4;
	    }
	  break;

	case '\b': *buffer_ptr++ = '\\'; *buffer_ptr++ = 'b';	break;
	case '\f': *buffer_ptr++ = '\\'; *buffer_ptr++ = 'f';	break;
	case '\n': *buffer_ptr++ = '\\'; *buffer_ptr++ = 'n';	break;
	case '\r': *buffer_ptr++ = '\\'; *buffer_ptr++ = 'r';	break;
	case '\t': *buffer_ptr++ = '\\'; *buffer_ptr++ = 't';	break;
	case '\v': *buffer_ptr++ = '\\'; *buffer_ptr++ = 'v';	break;
	case '\"': *buffer_ptr++ = '\\'; *buffer_ptr++ = '"';	break;
	case '\'': *buffer_ptr++ = '\\'; *buffer_ptr++ = '\'';	break;
	case '\\': *buffer_ptr++ = '\\'; *buffer_ptr++ = '\\';	break;
	}
    }

  buffer_ptr[0] = '"';
  buffer_ptr[1] = '\0';

  return print_buffer;
}


/* Convert a stab type into a string.  */

static const char *
printable_stab_type (int type)
{
  switch (type)
    {
      case 0x20: return "GSYM,  ";	/* global symbol: name,,0,type,0 */
      case 0x22: return "FNAME, ";	/* procedure name (f77 kludge): name,,0 */
      case 0x24: return "FUN,   ";	/* procedure: name,,0,linenumber,address */
      case 0x26: return "STSYM, ";	/* static symbol: name,,0,type,address */
      case 0x28: return "LCSYM, ";	/* .lcomm symbol: name,,0,type,address */
      case 0x2a: return "MAIN,  ";	/* Name of main routine, not used in C */
      case 0x30: return "PC,    ";	/* global pascal symbol: name,,0,subtype,line */
      case 0x40: return "RSYM,  ";	/* register sym: name,,0,type,register */
      case 0x44: return "SLINE, ";	/* src line: 0,,0,linenumber,address */
      case 0x46: return "DSLINE,";	/* like N_SLINE, except in data segment */
      case 0x48: return "BROWS, ";	/* source code browser stabs */
      case 0x54: return "CATCH, ";	/* GNU G++ catch clause */
      case 0x60: return "SSYM,  ";	/* structure elt: name,,0,type,struct_offset */
      case 0x64: return "SO,    ";	/* source file name: name,,0,0,address */
      case 0x80: return "LSYM,  ";	/* local sym: name,,0,type,offset */
      case 0x82: return "BINCL, ";	/* aux beginning of include file */
      case 0x84: return "SOL,   ";	/* #included file name: name,,0,0,address */
      case 0xa0: return "PSYM,  ";	/* parameter: name,,0,type,offset */
      case 0xa2: return "EINCL, ";	/* aux end of include file */
      case 0xa4: return "ENTRY, ";	/* alternate entry: name,linenumber,address */
      case 0xc0: return "LBRAC, ";	/* left bracket: 0,,0,nesting level,address */
      case 0xc2: return "EXCL,  ";	/* placeholder for deleted include file */
      case 0xe0: return "RBRAC, ";	/* right bracket: 0,,0,nesting level,address */
      case 0xe2: return "BCOMM, ";	/* begin common: name,, */
      case 0xe4: return "ECOMM, ";	/* end common: name,, */
      case 0xe8: return "ECOML, ";	/* end common (local name): ,,address */
      case 0xfe: return "LENG,  ";	/* second stab entry with length information */
    }

  return stringify (type, 2, 'x', 0, "", ",");
}


/* Return a section's name.  */

static const char *
get_section_name (Elf32_Word sec_index)
{
  Elf32_Shdr *sec_hdr;

  if (sec_index == 0)
    return "<dummy>";

  if (sec_index >= ehdr.e_shnum || ignore_string)
    return stringify (sec_index, 0, 'd', 0, "", "");

  sec_hdr = SEC_PTR (sec_index);
  return elf_string ((Elf32_Word)ehdr.e_shstrndx, sec_hdr->sh_name);
}


/* Print the section body by invoking objdump -d on it.  If objdump can't
   be found, fall back to V.4's dis.  */

static void
print_section_objdump (Elf32_Shdr *sec_hdr)
{
  pid_t pid;

  fflush (stdout);
  fflush (stderr);

  pid = fork ();
  if (pid == -1)		/* error, no fork */
    perror ("fork");

  else if (pid == 0)		/* child context */
    {
      const char *argv[6];

      argv[0] = "objdump";
      argv[1] = "-d";
      argv[2] = "-j";
      argv[3] = elf_string ((Elf32_Word)ehdr.e_shstrndx, sec_hdr->sh_name);
      argv[4] = file;
      argv[5] = (char *)0;

      execv ("../binutils/objdump", (char *const *) argv);
      execv ("../../binutils/objdump", (char *const *) argv);
      execvp ("objdump", (char *const *) argv);
      execv ("/usr/latest/bin/objdump", (char *const *) argv);
      execv ("/usr/ccs/gcc/objdump", (char *const *) argv);

      argv[0] = "dis";
      argv[1] = "-t";
      argv[2] = elf_string ((Elf32_Word)ehdr.e_shstrndx, sec_hdr->sh_name);
      argv[3] = file;
      argv[4] = (char *)0;
      execvp ("dis", (char *const *) argv);
      execv ("/usr/ccs/bin/dis", (char *const *) argv);

      perror ("objdump/dis");
      fflush (stderr);
      _exit (1);
    }

  else				/* parent context */
    {
      int status = 0;
      void (*sigint)(int)  = signal (SIGINT,  SIG_IGN);
      void (*sigquit)(int) = signal (SIGQUIT, SIG_IGN);

      if (waitpid (pid, &status, 0) < 0)
	perror ("waitpid");

      signal (SIGINT,  sigint);
      signal (SIGQUIT, sigquit);

      if (status)
	{
	  if (WIFSIGNALED (status))
	    fprintf (stderr, "objdump exited with signal %s\n",
		     stringify (WTERMSIG (status), 0, 'd', 0, "", ""));
	  else
	    fprintf (stderr, "objdump exited with %s\n",
		     stringify (WEXITSTATUS (status), 0, 'd', 0, "", ""));

	  fflush (stderr);
	}
    }
}


/* Print the section body just as a series of integers.  */

static void
print_section_raw (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	 size	   = sec_hdr->sh_size;
  Uchar		 *ptr	   = file_bytes + sec_hdr->sh_offset;
  int		 sec_off   = 0;
  int		 width_off = stringify_len (size-1, 'd', 1);
  Elf32_Word	 value;

  union { Elf32_Word i; char c[4]; } u;

  if (objdump && (sec_hdr->sh_flags & SHF_EXECINSTR))
    {
      print_section_objdump (sec_hdr);
      return;
    }

  while (size > 0)
    {
      if (size >= 4)
	{
	  if ((((unsigned long)ptr) & ~3u) == 0)
	    value = decode_elf32_word (*(Elf32_Word *)ptr);

	  else
	    {
	      memcpy (u.c, ptr, 4);
	      value = decode_elf32_word (u.i);
	    }

	  printf ("    Offset %s: 0x%.8lx  %11ld  %s\n",
		  stringify (sec_off, width_off, 'd', 0, "", ""),
		  (long)value,
		  (long)value,
		  printable_string (ptr, 4));

	  sec_off += 4;
	  ptr += 4;
	  size -= 4;
	}
      else
	{
	  value = *ptr;
	  printf ("    Offset %s:       0x%.2x  %11d  %s\n",
		  stringify (sec_off, width_off, 'd', 0, "", ""),
		  (int)value,
		  (int)value,
		  printable_string (ptr, 1));

	  sec_off++;
	  ptr++;
	  size--;
	}
    }
}


/* Print the special .stab section.  */

static void
print_section_stab (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	 size		= sec_hdr->sh_size;
  elf_stab_t	*ptr		= (elf_stab_t *)(file_bytes + sec_hdr->sh_offset);
  elf_stab_t	*endptr		= (elf_stab_t *)((char *)ptr + size);
  int		 width_soffset	= 0;
  int		 width_type	= 0;
  int		 width_other	= 0;
  int		 width_desc	= 0;
  int		 width_value	= 0;
  int		 width_stab_num	= stringify_len (endptr - ptr, 'd', 1);
  int		 stab_num	= 0;
  Elf32_Word	 string_sect	= 0;
  int		 len;
  Elf32_Word	 i;

  /* Look for .stabstr section */
  for (i = 0; i < ehdr.e_shnum; i++)
    {
      if (strcmp (get_section_name (i), ".stabstr") == 0)
	{
	  string_sect = i;
	  break;
	}
    }

  /* Find lengths for everything.  */
  while (ptr < endptr)
    {
      len = stringify_len (decode_elf32_off (ptr->soffset), 'd', 1);
      if (len > width_soffset)
	width_soffset = len;

      len = strlen (printable_stab_type (ptr->type));
      if (len > width_type)
	width_type = len;

      len = stringify_len (decode_elf32_half (ptr->desc), 'd', 1);
      if (len > width_desc)
	width_desc = len;

      len = stringify_len (decode_elf32_sword (ptr->value), 'd', 1);
      if (len > width_value)
	width_value = len;

      if (ptr->other)
	{
	  len = stringify_len (ptr->other, 'd', 1);
	  if (len > width_other)
	    width_other = len;
	}

      ptr++;
    }

  /* Print fields out.  */
  ptr = (elf_stab_t *)(file_bytes + sec_hdr->sh_offset);
  while (ptr < endptr)
    {
      if (width_other)
	printf ("    Stab #%s: type= %*s other= %s, desc= %s, value= %s, str offset= %s",
		stringify (++stab_num, width_stab_num, 'd', 0, "", ""),
		width_type, printable_stab_type (ptr->type),
		stringify (ptr->other, width_other, 'd', 0, "", ""),
		stringify (decode_elf32_half (ptr->desc), width_desc, 'd', 0, "", ""),
		stringify (decode_elf32_sword (ptr->value), width_value, 'd', 0, "", ""),
		stringify (decode_elf32_off (ptr->soffset), width_soffset, 'd', 0, "", ""));
      else
	printf ("    Stab #%s: type= %*s desc= %s, value= %s, str offset= %s",
		stringify (++stab_num, width_stab_num, 'd', 0, "", ""),
		width_type, printable_stab_type (ptr->type),
		stringify (decode_elf32_half (ptr->desc), width_desc, 'd', 0, "", ""),
		stringify (decode_elf32_sword (ptr->value), width_value, 'd', 0, "", ""),
		stringify (decode_elf32_off (ptr->soffset), width_soffset, 'd', 0, "", ""));

      if (string_sect)
	printf (", str= %s\n", elf_string (string_sect, decode_elf32_off (ptr->soffset)));
      else
	printf ("\n");

      ptr++;
    }
}
  

/* Print the symbol table entries.  */

static void
print_section_symtab (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	  size	 	= sec_hdr->sh_size;
  Elf32_Sym	 *ptr	 	= (Elf32_Sym *)(file_bytes + sec_hdr->sh_offset);
  Elf32_Sym	 *endptr 	= (Elf32_Sym *)((char *)ptr + size);
  Elf32_Sym	  sym;
  int		  sym_num	= 0;
  int		  width_value	= 0;
  int		  width_size	= 0;
  int		  width_shndx	= 0;
  int		  width_sym_num;
  int		  len;
  const char	 *type;
  const char	 *bind;
  const char	 *shndx;

  printf ("    String table = \"%s\"\n\n", get_section_name (sec_hdr->sh_link));

  /* Calculate minimum lengths */
  width_sym_num = stringify_len (endptr - ptr, 'd', 1);
  while (ptr < endptr)
    {
      decode_elf32_sym (&sym, ptr);
      len = stringify_len (sym.st_value, 'x', 1);
      if (width_value < len)
	width_value = len;

      len = stringify_len (sym.st_size, 'd', 1);
      if (width_size < len)
	width_size = len;

      switch (sym.st_shndx)
	{
	default:
	  len = (ignore_string)
			? stringify_len (sym.st_shndx, 'd', 0)
			: strlen (get_section_name (sym.st_shndx));
	  break;

	case SHN_UNDEF:  len = sizeof ("undef")-1;	break;
	case SHN_ABS:    len = sizeof ("abs")-1;	break;
	case SHN_COMMON: len = sizeof ("common")-1;	break;
	}

      if (width_shndx < len)
	width_shndx = len;

      ptr++;
    }

  ptr = (Elf32_Sym *)(file_bytes + sec_hdr->sh_offset);
  while (ptr < endptr)
    {
      decode_elf32_sym (&sym, ptr);
      switch (ELF32_ST_BIND (sym.st_info))
	{
	default:
	  bind = stringify (ELF32_ST_BIND (sym.st_info), 0, 'd', 0, "", ",");
	  break;

	case STB_LOCAL:  bind = "local,";	break;
	case STB_GLOBAL: bind = "global,";	break;
	case STB_WEAK:   bind = "weak,"; 	break;
	}

      switch (ELF32_ST_TYPE (sym.st_info))
	{
	default:
	  type = stringify (ELF32_ST_TYPE (sym.st_info), 0, 'd', 0, "", ",");
	  break;

	case STT_NOTYPE:	type = "none,";		break;
	case STT_OBJECT:	type = "obj,";		break;
	case STT_FUNC:		type = "func,";		break;
	case STT_SECTION:	type = "sect,";		break;
	case STT_FILE:		type = "file,";		break;
	}

      switch ((int)sym.st_shndx)
	{
	default:
	  shndx = (ignore_string)
			? stringify (sym.st_shndx, 0, 'd', 0, "", "")
			: get_section_name (sym.st_shndx);
	  break;

	case (int)SHN_UNDEF:	shndx = "undef";	break;
	case (int)SHN_ABS:	shndx = "abs";		break;
	case (int)SHN_COMMON:	shndx = "common";	break;
	}

      printf ("    Sym %s: value= %s, size= %s sect= %s%-*s type= %-5s bind= %-7s name= %s\n",
	      stringify (sym_num, width_sym_num, 'd', 0, "", ""),
	      stringify (sym.st_value, width_value, 'x', 0, "", ""),
	      stringify (sym.st_size, width_size, 'd', 0, "", ""),
	      shndx, width_shndx + 1 - (int)strlen (shndx), ",",
	      type, bind, elf_string (sec_hdr->sh_link, sym.st_name));

      ptr++;
      sym_num++;
    }
}


/* Print the hash table.  */

static void
print_section_hash (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	 size	 = sec_hdr->sh_size;
  Elf32_Word	*ptr	 = (Elf32_Word *)(file_bytes + sec_hdr->sh_offset);
  Elf32_Word	*endptr	 = (Elf32_Word *)((char *)ptr + size);
  Elf32_Word	 nbucket = decode_elf32_word (*ptr++);
  Elf32_Word	 nchain  = decode_elf32_word (*ptr++);
  Elf32_Word	*hchain  = ptr + nbucket;
  int		 loading;
  Elf32_Word	 i;
  Elf32_Word	 element;
  Elf32_Word	 next;
  int		 width;


  if (endptr != (ptr + nbucket + nchain))
    {
      printf ("    <hash table wrong size>\n");
      print_section_raw (sec_hdr);
      return;
    }

  width = stringify_len ((nchain > nbucket) ? nchain : nbucket, 'd', 1);
  printf ("    Symbol table = \"%s\"\n", get_section_name (sec_hdr->sh_link));
  printf ("    # Buckets    = %s\n", stringify (nbucket, width, 'd', 0, "", ""));
  printf ("    # Elements   = %s\n\n", stringify (nchain, width, 'd', 0, "", ""));

  for (i = 0; i < nbucket; i++)
    {
      loading = 1;
      element = decode_elf32_word (*ptr);
      while ((next = decode_elf32_word (hchain[element])) != 0)
	{
	  loading++;
	  element = next;
	}

      printf ("    Bucket %s: first: %s, loading: %s\n",
	      stringify (i, width, 'd', 0, "", ""),
	      stringify (decode_elf32_word (*ptr++), width, 'd', 0, "", ""),
	      stringify (loading, width, 'd', 0, "", ""));
    }

  putchar ('\n');
  for (i = 0; i < nchain; i++)
    printf ("    Chain  %s: %s\n",
	    stringify (i, width, 'd', 0, "", ""),
	    stringify (decode_elf32_word (*ptr++), width, 'd', 0, "", ""));
}


/* Print the string section body.  */

static void
print_section_string (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	  size	    = sec_hdr->sh_size;
  char		 *ptr	    = (char *)(file_bytes + sec_hdr->sh_offset);
  char		 *endptr    = ptr + size;
  int		  sec_off   = 0;
  int		  len	    = 0;
  int		  width_off = stringify_len (size-1, 'd', 1);

  while (ptr < endptr)
    {
      if (*ptr != '\0' || sec_off == 0)
	printf ("    Offset %s: \"%s\"\n", stringify (sec_off, width_off, 'd', 0, "", ""), ptr);

      len = strlen (ptr) + 1;
      ptr += len;
      sec_off += len;
    }
}


/* Decode a relocation type into a printable value.  */

static const char *
get_reloc_type (int type)
{
  struct reloc_lookup {
    int		reloc;			/* relocation number */
    const char *string;			/* printable string */
  };

  int machine = ehdr.e_machine;
  const struct reloc_lookup *ptr;

  static const struct reloc_lookup lookup_i386[] =
    { /* Reloc */	/* String */
      {  0,		"R_386_NONE," },
      {  1,		"R_386_32," },
      {  2,		"R_386_PC32," },
      {  3,		"R_386_GOT," },
      {  4,		"R_386_PLT32," },
      {  5,		"R_386_COPY," },
      {  6,		"R_386_GLOB_DAT," },
      {  7,		"R_386_JMP_SLOT," },
      {  8,		"R_386_RELATIVE," },
      {  9,		"R_386_GOTOFF," },
      { 10,		"R_386_GOTPC," },
      {  0,		(char *)0 },
    };

  static const struct reloc_lookup lookup_sparc[] =
    { /* Reloc */	/* String */
	{ 0, "R_SPARC_NONE" },
	{ 1, "R_SPARC_8" },
	{ 2, "R_SPARC_16" },
	{ 3, "R_SPARC_32" },
	{ 4, "R_SPARC_DISP8" },
	{ 5, "R_SPARC_DISP16" },
	{ 6, "R_SPARC_DISP32" },
	{ 7, "R_SPARC_WDISP30" },
	{ 8, "R_SPARC_WDISP22" },
	{ 9, "R_SPARC_HI22" },
	{ 10, "R_SPARC_22" },
	{ 11, "R_SPARC_13" },
	{ 12, "R_SPARC_LO10" },
	{ 13, "R_SPARC_GOT10" },
	{ 14, "R_SPARC_GOT13" },
	{ 15, "R_SPARC_GOT22" },
	{ 16, "R_SPARC_PC10" },
	{ 17, "R_SPARC_PC22" },
	{ 18, "R_SPARC_WPLT30" },
	{ 19, "R_SPARC_COPY" },
	{ 20, "R_SPARC_GLOB_DAT" },
	{ 21, "R_SPARC_JMP_SLOT" },
	{ 22, "R_SPARC_RELATIVE" },
	{ 23, "R_SPARC_UA32" },
	{ 24, "R_SPARC_NUM" },
	{ 0, 0 },
    };

  static const struct reloc_lookup lookup_ppc[] =
    { /* Reloc */	/* String */
      {   0,		"R_PPC_NONE" },
      {   1,		"R_PPC_ADDR32" },
      {   2,		"R_PPC_ADDR24" },
      {   3,		"R_PPC_ADDR16" },
      {   4,		"R_PPC_ADDR16_LO" },
      {   5,		"R_PPC_ADDR16_HI" },
      {   6,		"R_PPC_ADDR16_HA" },
      {   7,		"R_PPC_ADDR14" },
      {   8,		"R_PPC_ADDR14_BRTAKEN" },
      {   9,		"R_PPC_ADDR14_BRNTAKEN" },
      {  10,		"R_PPC_REL24" },
      {  11,		"R_PPC_REL14" },
      {  12,		"R_PPC_REL14_BRTAKEN" },
      {  13,		"R_PPC_REL14_BRNTAKEN" },
      {  14,		"R_PPC_GOT16" },
      {  15,		"R_PPC_GOT16_LO" },
      {  16,		"R_PPC_GOT16_HI" },
      {  17,		"R_PPC_GOT16_HA" },
      {  18,		"R_PPC_PLT24" },
      {  19,		"R_PPC_COPY" },
      {  21,		"R_PPC_JMP_SLOT" },
      {  22,		"R_PPC_RELATIVE" },
      {  23,		"R_PPC_LOCAL24PC" },
      {  24,		"R_PPC_UADDR32" },
      {  25,		"R_PPC_UADDR16" },
      {  26,		"R_PPC_REL32" },
      {  27,		"R_PPC_PLT32" },
      {  28,		"R_PPC_PLTREL32" },
      {  29,		"R_PPC_PLT16_LO" },
      {  30,		"R_PPC_PLT16_HI" },
      {  31,		"R_PPC_PLT16_HA" },
      {  32,		"R_PPC_SDAREL" },
      {  33,		"R_PPC_SECTOFF" },
      {  34,		"R_PPC_SECTOFF_LO" },
      {  35,		"R_PPC_SECTOFF_HI" },
      {  36,		"R_PPC_SECTOFF_HA" },
      { 101,		"R_PPC_EMB_NADDR32" },
      { 102,		"R_PPC_EMB_NADDR16" },
      { 103,		"R_PPC_EMB_NADDR16_LO" },
      { 104,		"R_PPC_EMB_NADDR16_HI" },
      { 105,		"R_PPC_EMB_NADDR16_HA" },
      { 106,		"R_PPC_EMB_SDAI16" },
      { 107,		"R_PPC_EMB_SDA2I16" },
      { 108,		"R_PPC_EMB_SDA2REL" },
      { 109,		"R_PPC_EMB_SDA21" },
      { 110,		"R_PPC_EMB_MRKREF" },
      { 111,		"R_PPC_EMB_RELSEC16" },
      { 112,		"R_PPC_EMB_RELST_LO" },
      { 113,		"R_PPC_EMB_RELST_HI" },
      { 114,		"R_PPC_EMB_RELST_HA" },
      { 115,		"R_PPC_EMB_BIT_FLD" },
      { 116,		"R_PPC_EMB_RELSDA" },
      {   0,		(char *)0 },
    };

  switch (machine)
    {
    default:
      ptr = (struct reloc_lookup *)0;
      break;

    case EM_386:
      ptr = &lookup_i386[0];
      break;

    case EM_SPARC:
      ptr = &lookup_sparc[0];
      break;
	
    case EM_PPC:
    case EM_CYGNUS_POWERPC:
    case EM_PPC_OLD:
      ptr = &lookup_ppc[0];
      break;
    }

  if (ptr)
    {
      for ( ; ptr->string != (char *)0; ptr++)
	{
	  if (type == ptr->reloc)
	    return ptr->string;
	}
    }

  return stringify (type, 0, 'd', 0, "", ",");
}


/* Print RELA relocations.  */

static void
print_section_rela (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	 size		= sec_hdr->sh_size;
  Elf32_Rela	*ptr		= (Elf32_Rela *)(file_bytes + sec_hdr->sh_offset);
  Elf32_Rela	*endptr		= (Elf32_Rela *)((char *)ptr + size);
  Elf32_Shdr	*sym_sec	= SEC_PTR (sec_hdr->sh_link);
  Elf32_Sym	*sym_ptr	= (Elf32_Sym *)(file_bytes + sym_sec->sh_offset);
  int		 rel_object_p	= (ehdr.e_type == ET_REL);
  int		 rel_num	= 0;
  int		 width_offset	= 0;
  int		 width_info	= 0;
  int		 width_sym	= 0;
  int		 width_type	= 0;
  int		 width_addend	= 0;
  int		 width_rel_num;
  int		 sym_num;
  int		 len;
  Elf32_Rela	 rela;
  Elf32_Word	 offset;
  const char	*sect_or_undef;
  const char	*name;
  const char	*offset_value;

  printf ("    Symbol table section    = \"%s\"\n",   get_section_name (sec_hdr->sh_link));
  printf ("    Section being relocated = \"%s\"\n\n", get_section_name (sec_hdr->sh_info));

  /* Calculate minimum lengths */
  width_rel_num = stringify_len (endptr - ptr, 'd', 1);
  while (ptr < endptr)
    {
      decode_elf32_rela (&rela, ptr);
      if (rel_object_p)
	offset = rela.r_offset;
      else
	offset = rela.r_offset - sec_hdr->sh_addr;

      len = stringify_len (offset, 'd', 1);
      if (width_offset < len)
	width_offset = len;

      len = stringify_len (rela.r_info, 'd', 1);
      if (width_info < len)
	width_info = len;

      len = stringify_len (ELF32_R_SYM (rela.r_info), 'd', 1);
      if (width_sym < len)
	width_sym = len;

      len = strlen (get_reloc_type (ELF32_R_TYPE (rela.r_info)));
      if (width_type < len)
	width_type = len;

      len = stringify_len (rela.r_addend, 'd', 0);
      if (width_addend < len)
	width_addend = len;

      ptr++;
    }

  ptr = (Elf32_Rela *)(file_bytes + sec_hdr->sh_offset);
  while (ptr < endptr)
    {
      decode_elf32_rela (&rela, ptr);
      if (rel_object_p)
	{
	  offset = rela.r_offset;
	  offset_value = "";
	}
      else
	{
	  offset = rela.r_offset - sec_hdr->sh_addr;
	  offset_value = stringify (rela.r_offset, 8, 'x', 0, " [", "]");
	}

      sym_num = ELF32_R_SYM (rela.r_info);
      switch ((int) sym_ptr[sym_num].st_shndx)
	{
	default:		sect_or_undef = "internal";	break;
	case (int)SHN_UNDEF:	sect_or_undef = "external";	break;
	case (int)SHN_COMMON:	sect_or_undef = "common  ";	break;
	case (int)SHN_ABS:	sect_or_undef = "absolute";	break;
	}

      name = elf_string (sym_sec->sh_link, sym_ptr[sym_num].st_name);
      printf ("    Reloc %s: offset= %s%s, info= %s, sym= %s, type= %-*s addend= %s, %s \"%s\"\n",
	      stringify (rel_num, width_rel_num, 'd', 0, "", ""),
	      stringify (offset, width_offset, 'd', 0, "", ""), offset_value,
	      stringify (rela.r_info, width_info, 'd', 0, "", ""),
	      stringify (sym_num, width_sym, 'd', 0, "", ""),
	      width_type, get_reloc_type (ELF32_R_TYPE (rela.r_info)),
	      stringify (rela.r_addend, width_addend, 'd', 0, "", ""),
	      sect_or_undef, name);

      ptr++;
      rel_num++;
    }
}


/* Print REL relocations.  */

static void
print_section_rel (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	 size		= sec_hdr->sh_size;
  Elf32_Rel	*ptr		= (Elf32_Rel *)(file_bytes + sec_hdr->sh_offset);
  Elf32_Rel	*endptr		= (Elf32_Rel *)((char *)ptr + size);
  Elf32_Shdr	*sym_sec	= SEC_PTR (sec_hdr->sh_link);
  Elf32_Sym	*sym_ptr	= (Elf32_Sym *)(file_bytes + sym_sec->sh_offset);
  Elf32_Shdr	*data_sec	= SEC_PTR (sec_hdr->sh_info);
  char		*data_ptr	= (char *)(file_bytes + data_sec->sh_offset);
  int		 rel_object_p	= (ehdr.e_type == ET_REL);
  int		 rel_num	= 0;
  int		 width_offset	= 0;
  int		 width_info	= 0;
  int		 width_sym	= 0;
  int		 width_type	= 0;
  int		 width_addend	= 0;
  int		 width_rel_num;
  int		 sym_num;
  int		 len;
  Elf32_Rel	 rel;
  Elf32_Word	 offset;
  const char	*sect_or_undef;
  const char	*name;
  const char	*offset_value;
  const char	*addend_value;

  union {
    Elf32_Word	i;
    char	c[sizeof (Elf32_Word)];
  } addend;

  printf ("    Symbol table section    = \"%s\"\n",   get_section_name (sec_hdr->sh_link));
  printf ("    Section being relocated = \"%s\"\n\n", get_section_name (sec_hdr->sh_info));

  /* Calculate minimum lengths */
  width_rel_num = stringify_len (endptr - ptr, 'd', 1);
  while (ptr < endptr)
    {
      decode_elf32_rel (&rel, ptr);
      if (rel_object_p)
	offset = ptr->r_offset;
      else
	offset = ptr->r_offset - sec_hdr->sh_addr;

      len = stringify_len (offset, 'd', 1);
      if (width_offset < len)
	width_offset = len;

      len = stringify_len (ptr->r_info, 'd', 1);
      if (width_info < len)
	width_info = len;

      len = stringify_len (ELF32_R_SYM (ptr->r_info), 'd', 1);
      if (width_sym < len)
	width_sym = len;

      len = strlen (get_reloc_type (ELF32_R_TYPE (ptr->r_info)));
      if (width_type < len)
	width_type = len;

      if (offset < data_sec->sh_size)
	{
	  memcpy (addend.c, data_ptr + offset, sizeof (Elf32_Word));
	  len = stringify_len (addend.i, 'd', 0);
	}
      else
	len = sizeof ("error")-1;

      if (width_addend < len)
	width_addend = len;

      ptr++;
    }

  ptr = (Elf32_Rel *)(file_bytes + sec_hdr->sh_offset);
  while (ptr < endptr)
    {
      decode_elf32_rel (&rel, ptr);
      if (rel_object_p)
	{
	  offset = rel.r_offset;
	  offset_value = "";
	}
      else
	{
	  offset = rel.r_offset - sec_hdr->sh_addr;
	  offset_value = stringify (rel.r_offset, 8, 'x', 0, " [", "]");
	}

      if (offset < data_sec->sh_size)
	{
	  memcpy (addend.c, data_ptr + offset, sizeof (Elf32_Word));
	  addend_value = stringify (addend.i, 0, 'd', 0, "", "");
	}
      else
	addend_value = "error";

      sym_num = ELF32_R_SYM (rel.r_info);
      switch ((int)sym_ptr[sym_num].st_shndx)
	{
	default:		sect_or_undef = "internal";	break;
	case (int)SHN_UNDEF:	sect_or_undef = "external";	break;
	case (int)SHN_COMMON:	sect_or_undef = "common  ";	break;
	case (int)SHN_ABS:	sect_or_undef = "absolute";	break;
	}

      name = elf_string (sym_sec->sh_link, sym_ptr[sym_num].st_name);
      printf ("    Reloc %s: offset= %s%s, info= %s, sym= %s, type= %-*s addend= %*s, %s \"%s\"\n",
	      stringify (rel_num, width_rel_num, 'd', 0, "", ""),
	      stringify (offset, width_offset, 'd', 0, "", ""), offset_value,
	      stringify (rel.r_info, width_info, 'd', 0, "", ""),
	      stringify (sym_num, width_sym, 'd', 0, "", ""),
	      width_type, get_reloc_type (ELF32_R_TYPE (rel.r_info)),
	      width_addend, addend_value,
	      sect_or_undef, name);

      ptr++;
      rel_num++;
    }
}


/* Print the dynamic link structures.  */

static void
print_section_dynamic (Elf32_Shdr *sec_hdr)
{
  Elf32_Word	 size		= sec_hdr->sh_size;
  Elf32_Dyn	*ptr		= (Elf32_Dyn *)(file_bytes + sec_hdr->sh_offset);
  Elf32_Dyn	*endptr		= (Elf32_Dyn *)((char *)ptr + size);
  int		 dyn_num	= 0;
  int		 width_tag_str	= 0;
  int		 width_val_num	= 0;
  int		 width_dyn_num;
  int		 len;
  const char	*tag;
  Elf32_Dyn	 dyn;

  /* Calculate minimum lengths */
  width_dyn_num = stringify_len (endptr - ptr, 'd', 1);
  while (ptr < endptr)
    {
      decode_elf32_dyn (&dyn, ptr);
      if (dyn.d_tag >= 0 && (size_t)dyn.d_tag < sizeof (dyn_tag) / sizeof (dyn_tag[0]))
	len = strlen (dyn_tag[ dyn.d_tag ]);
      else
	len = stringify_len (dyn.d_tag, 'd', 1);

      if (width_tag_str < len)
	width_tag_str = len;

      len = stringify_len (dyn.d_un.d_val, 'd', 1);
      if (width_val_num < len)
	width_val_num = len;

      ptr++;
    }

  ptr = (Elf32_Dyn *)(file_bytes + sec_hdr->sh_offset);
  while (ptr < endptr)
    {
      decode_elf32_dyn (&dyn, ptr);
      if (dyn.d_tag >= 0 && (size_t)dyn.d_tag < sizeof (dyn_tag) / sizeof (dyn_tag[0]))
	tag = dyn_tag[ dyn.d_tag ];
      else
	tag = stringify (dyn.d_tag, 0, 'd', 0, "", "");

      printf ("    Dynamic record %s: tag= %*s, value= 0x%.8lx, %*ld\n",
	      stringify (dyn_num, width_dyn_num, 'd', 0, "", ""),
	      width_tag_str, tag,
	      (long)dyn.d_un.d_val, width_val_num, (long)dyn.d_un.d_val);

      ptr++;
      dyn_num++;
    }
}


/* Print the progbits section.  */

static void
print_section_progbits (Elf32_Shdr *sec_hdr)
{
  const char *name = elf_string ((Elf32_Word)ehdr.e_shstrndx, sec_hdr->sh_name);

  if (strcmp (name, ".interp") == 0)
    print_section_string (sec_hdr);

  else if (no_progbits)
    printf ("    Section body skipped...\n");

  else if (strcmp (name, ".comment") == 0)
    print_section_string (sec_hdr);

  else if (strcmp (name, ".stab") == 0)
    print_section_stab (sec_hdr);

  else if (strcmp (name, ".stabstr") == 0)
    print_section_string (sec_hdr);

  else if (strcmp (name, ".debug_sfnames") == 0)
    print_section_string (sec_hdr);

  else
    print_section_raw (sec_hdr);
}


/* Size the max fields in printing section headers. */

static void
size_section (Elf32_Word sec_index)
{
  Elf32_Shdr	  *sec_hdr	= SEC_PTR (sec_index);
  const char	  *name;
  const char	  *sec_type;
  struct sec_list *sec_ptr;
  int		   len;

  name = get_section_name (sec_index);
  if (sec_index)		/* skip dummy section */
    {
      add_layout (&file_layout, name, sec_hdr->sh_offset,
		  ((int)sec_hdr->sh_type == SHT_NOBITS) ? 0 : sec_hdr->sh_size, layout_free_no);

      if (ehdr.e_phnum && (sec_hdr->sh_flags & SHF_ALLOC))
	add_layout (&vm_layout, name, sec_hdr->sh_addr, sec_hdr->sh_size, layout_free_no);
    }

  if (!summary)
    return;

  /* Skip section if desired.  */
  if (sections)
    {
      for (sec_ptr = sections; sec_ptr != (struct sec_list *)0; sec_ptr = sec_ptr->next)
	{
	  if (strcmp (name, sec_ptr->section) == 0)
	    break;
	}

      if (!sec_ptr)
	return;
    }

  len = stringify_len (sec_index, 'd', 1);
  if (len > sh_index_width)
    sh_index_width = len;

  len = strlen (name);
  if (len > sh_name_width)
    sh_name_width = len;

  switch ((int)sec_hdr->sh_type)
    {
    default:
      sec_type = stringify (sec_hdr->sh_type, 0, 'd', 0, "type=", "");
      break;

    case SHT_NULL:	sec_type = "none";	break;
    case SHT_PROGBITS:	sec_type = "progbits";	break;
    case SHT_SYMTAB:	sec_type = "symtab";	break;
    case SHT_STRTAB:	sec_type = "strtab";	break;
    case SHT_RELA:	sec_type = "rela";	break;
    case SHT_HASH:	sec_type = "hash";	break;
    case SHT_DYNAMIC:	sec_type = "dynamic";	break;
    case SHT_NOTE:	sec_type = "note";	break;
    case SHT_NOBITS:	sec_type = "nobits";	break;
    case SHT_REL:	sec_type = "rel";	break;
    case SHT_SHLIB:	sec_type = "shlib";	break;
    case SHT_DYNSYM:	sec_type = "dynsym";	break;
    }

  len = strlen (sec_type);
  if (len > sh_type_width)
    sh_type_width = len;

  len = stringify_len (sec_hdr->sh_offset, 'd', 1);
  if (len > sh_start_width)
    sh_start_width = len;

  len = stringify_len (sec_hdr->sh_size, 'd', 1);
  if (len > sh_size_width)
    sh_size_width = len;

  len = stringify_len (sec_hdr->sh_offset + (((int)sec_hdr->sh_type == SHT_NOBITS) ? 0 : sec_hdr->sh_size), 'x', 1);
  if (len > sh_end_width)
    sh_end_width = len;

  len = stringify_len (sec_hdr->sh_addr, 'x', 1);
  if (len > sh_addr_width)
    sh_addr_width = len;

  len = stringify_len (sec_hdr->sh_addralign, 'd', 1);
  if (len > sh_align_width)
    sh_align_width = len;

  len = stringify_len (sec_hdr->sh_flags, 'x', 1);
  if (len > sh_flags_width)
    sh_flags_width = len;

  len = stringify_len (sec_hdr->sh_info, 'd', 1);
  if (len > sh_info_width)
    sh_info_width = len;

  len = stringify_len (sec_hdr->sh_link, 'd', 1);
  if (len > sh_link_width)
    sh_link_width = len;

  len = stringify_len (sec_hdr->sh_entsize, 'd', 1);
  if (len > sh_esize_width)
    sh_esize_width = len;
}


/* Print a section header. */

static void
print_section (Elf32_Word sec_index)
{
  Elf32_Shdr	  *sec_hdr	= SEC_PTR (sec_index);
  const char	  *sec_type;
  const char	  *name;
  struct sec_list *sec_ptr;
  void (*sec_func)(Elf32_Shdr *);

  /* Skip section if desired.  */
  name = get_section_name (sec_index);
  if (sections)
    {
      for (sec_ptr = sections; sec_ptr != (struct sec_list *)0; sec_ptr = sec_ptr->next)
	{
	  if (strcmp (name, sec_ptr->section) == 0)
	    break;
	}

      if (!sec_ptr)
	return;
    }

  switch ((int)sec_hdr->sh_type)
    {
    default:
      sec_type = stringify (sec_hdr->sh_type, 0, 'd', 0, "type=", "");
      sec_func = print_section_raw;
      break;

    case SHT_NULL:	sec_type = "none";	sec_func = print_section_raw;		break;
    case SHT_PROGBITS:	sec_type = "progbits";	sec_func = print_section_progbits;	break;
    case SHT_SYMTAB:	sec_type = "symtab";	sec_func = print_section_symtab;	break;
    case SHT_STRTAB:	sec_type = "strtab";	sec_func = print_section_string;	break;
    case SHT_RELA:	sec_type = "rela";	sec_func = print_section_rela;		break;
    case SHT_HASH:	sec_type = "hash";	sec_func = print_section_hash;		break;
    case SHT_DYNAMIC:	sec_type = "dynamic";	sec_func = print_section_dynamic;	break;
    case SHT_NOTE:	sec_type = "note";	sec_func = print_section_raw;		break;
    case SHT_NOBITS:	sec_type = "nobits";	sec_func = print_section_raw;		break;
    case SHT_REL:	sec_type = "rel";	sec_func = print_section_rel;		break;
    case SHT_SHLIB:	sec_type = "shlib";	sec_func = print_section_raw;		break;
    case SHT_DYNSYM:	sec_type = "dynsym";	sec_func = print_section_symtab;	break;
    }

  printf ("%sSect %s %-*s %-*s off=%s %s size=%s addr=%s align=%s flag=%s [%s%s%s] esize=%s info=%s link=%s\n",
	  (!summary || sec_index == 0) ? "\n" : "",
	  stringify (sec_index, sh_index_width, 'd', 0, "", ""),
	  sh_name_width, name,
	  sh_type_width, sec_type,
	  stringify (sec_hdr->sh_offset, sh_start_width, 'x', 0, "", ""),
	  stringify (sec_hdr->sh_offset + ((sec_hdr->sh_type == SHT_NOBITS) ? 0 : sec_hdr->sh_size),
		     sh_end_width, 'x', 0, "", ""),
	  stringify (sec_hdr->sh_size, sh_size_width, 'd', 0, "", ""),
	  stringify (sec_hdr->sh_addr, sh_addr_width, 'x', 0, "", ""),
	  stringify (sec_hdr->sh_addralign, sh_align_width, 'd', 0, "", ""),
	  stringify (sec_hdr->sh_flags, sh_flags_width, 'x', 0, "", ""),
	  (sec_hdr->sh_flags & SHF_EXECINSTR)	? "x" : "-",
	  (sec_hdr->sh_flags & SHF_ALLOC)	? "a" : "-",
	  (sec_hdr->sh_flags & SHF_WRITE)	? "w" : "-",
	  stringify (sec_hdr->sh_entsize, sh_esize_width, 'd', 0, "", ""),
	  stringify (sec_hdr->sh_info, sh_info_width, 'd', 0, "", ""),
	  stringify (sec_hdr->sh_link, sh_link_width, 'd', 0, "", ""));

  if (sec_hdr->sh_size > 0 && sec_hdr->sh_type != SHT_NOBITS && !summary)
    {
      putchar ('\n');
      (*sec_func) (sec_hdr);
    }
}


/* Print program headers.  */

static void
print_proghdr (void)
{
  Elf32_Phdr *prog_hdr;
  const char *type;
  Elf32_Word ph_index;
  int width_index	= stringify_len (ehdr.e_phnum, 'd', 1);
  int width_vaddr	= 0;
  int width_vend	= 0;
  int width_type	= 0;
  int width_align	= 0;
  int width_flags	= 0;
  int width_offset	= 0;
  int width_offend	= 0;
  int width_filesz	= 0;
  int width_memsz	= 0;
  int len;
  char flags[4];
  char *load_start;
  char *load_end;
  const char *load_num;

  putchar ('\n');

  for (ph_index = 0; ph_index < ehdr.e_phnum; ph_index++)
    {
      prog_hdr = PROGHDR_PTR (ph_index);

      switch ((int)prog_hdr->p_type)
	{
	default:
	  type = stringify (prog_hdr->p_type, 0, 'd', 0, "type=", "");
	  break;

	case PT_NULL:	 type = "null,";	break;
	case PT_LOAD:	 type = "load,";	break;
	case PT_DYNAMIC: type = "dynamic,";	break;
	case PT_INTERP:	 type = "interp,";	break;
	case PT_NOTE:	 type = "note,";	break;
	case PT_SHLIB:	 type = "shlib,";	break;
	case PT_PHDR:	 type = "proghdr,";	break;
	}

      len = strlen (type);
      if (len > width_type)
	width_type = len;

      len = stringify_len (prog_hdr->p_vaddr, 'x', 1);
      if (len > width_vaddr)
	width_vaddr = len;

      len = stringify_len (prog_hdr->p_vaddr + prog_hdr->p_memsz, 'x', 1);
      if (len > width_vend)
	width_vend = len;

      len = stringify_len (prog_hdr->p_align, 'd', 1);
      if (len > width_align)
	width_align = len;

      len = stringify_len (prog_hdr->p_flags, 'x', 1);
      if (len > width_flags)
	width_flags = len;

      len = stringify_len (prog_hdr->p_offset, 'x', 1);
      if (len > width_offset)
	width_offset = len;

      len = stringify_len (prog_hdr->p_offset + prog_hdr->p_filesz, 'x', 1);
      if (len > width_offend)
	width_offend = len;

      len = stringify_len (prog_hdr->p_filesz, 'd', 1);
      if (len > width_filesz)
	width_filesz = len;

      len = stringify_len (prog_hdr->p_memsz, 'd', 1);
      if (len > width_memsz)
	width_memsz = len;
    }

  for (ph_index = 0; ph_index < ehdr.e_phnum; ph_index++)
    {
      prog_hdr = PROGHDR_PTR (ph_index);

      switch ((int)prog_hdr->p_type)
	{
	default:
	  type = stringify (prog_hdr->p_type, 0, 'd', 0, "type=", "");
	  break;

	case PT_NULL:	 type = "null,";	break;
	case PT_LOAD:	 type = "load,";	break;
	case PT_DYNAMIC: type = "dynamic,";	break;
	case PT_INTERP:	 type = "interp,";	break;
	case PT_NOTE:	 type = "note,";	break;
	case PT_SHLIB:	 type = "shlib,";	break;
	case PT_PHDR:	 type = "proghdr,";	break;
	}

      flags[0] = (prog_hdr->p_flags & PF_R) ? 'r' : '-';
      flags[1] = (prog_hdr->p_flags & PF_W) ? 'w' : '-';
      flags[2] = (prog_hdr->p_flags & PF_X) ? 'x' : '-';
      flags[3] = '\0';

      printf ("Hdr %s: %-*s vaddr= %s %s, align= %s, flags= %s [%s], off= %s %s, filesz= %s, memsz= %s\n",
	      stringify (ph_index, width_index, 'd', 0, "", ""),
	      width_type, type,
	      stringify (prog_hdr->p_vaddr, width_vaddr, 'x', 0, "", ""),
	      stringify (prog_hdr->p_vaddr + prog_hdr->p_memsz, width_vend, 'x', 0, "", ""),
	      stringify (prog_hdr->p_align, width_align, 'd', 0, "", ""),
	      stringify (prog_hdr->p_flags, width_flags, 'x', 0, "", ""),
	      flags,
	      stringify (prog_hdr->p_offset, width_offset, 'x', 0, "", ""),
	      stringify (prog_hdr->p_offset + prog_hdr->p_filesz, width_offend, 'x', 0, "", ""),
	      stringify (prog_hdr->p_filesz, width_filesz, 'd', 0, "", ""),
	      stringify (prog_hdr->p_memsz,  width_memsz,  'd', 0, "", ""));

      if ((int)prog_hdr->p_type == PT_LOAD)
	{
	  load_num = stringify (ph_index, 0, 'd', 0, "", "");
	  len = strlen (load_num) + strlen (flags);

	  load_start = (char *) xmalloc (sizeof("Load cmd #, region start []") + len);
	  sprintf (load_start, "Load cmd #%s, region start [%s]", load_num, flags);
	  add_layout (&vm_layout, load_start, prog_hdr->p_vaddr, 0, layout_free_no);
	  add_layout (&file_layout, load_start, prog_hdr->p_offset, 0, layout_free_yes);

	  load_end = (char *) xmalloc (sizeof("Load cmd #, region end []") + len);
	  sprintf (load_end, "Load cmd #%s, region end [%s]", load_num, flags);
	  add_layout (&vm_layout, load_end, prog_hdr->p_vaddr + prog_hdr->p_memsz, 0, layout_free_no);
	  add_layout (&file_layout, load_end, prog_hdr->p_offset + prog_hdr->p_filesz, 0, layout_free_yes);
	}
    }
}


/* Scan for linker generated symbols, and put them in the VM layout */

static void
find_linker_symbols (void)
{
  int		 i;
  Elf32_Shdr	*sec_hdr;
  const char	*name;
  Elf32_Sym	*ptr;
  Elf32_Sym	*endptr;
  Elf32_Sym	 sym;
  layout_t	*layout_ptr;

  for (i = 0; i < ehdr.e_shnum; i++)
    {
      sec_hdr = SEC_PTR (i);
      if ((int)sec_hdr->sh_type != SHT_SYMTAB && (int)sec_hdr->sh_type != SHT_DYNSYM)
	continue;

      ptr    = (Elf32_Sym *)(file_bytes + sec_hdr->sh_offset);
      endptr = (Elf32_Sym *)((char *)ptr + sec_hdr->sh_size);

      for ( ; ptr < endptr; ptr++)
	{
	  decode_elf32_sym (&sym, ptr);
	  name = elf_string (sec_hdr->sh_link, sym.st_name);
	  if (name[0] == '\0')
	    continue;

	  switch ((name[0] << 16) | (name[1] << 8) | name[2])
	    {
	    default:
	      continue;

	    case (('_' << 16) | ('G' << 8) | 'L'):
	      if (strcmp (name, "_GLOBAL_OFFSET_POINTER") != 0
		  && strcmp (name, "_GLOBAL_OFFSET_TABLE_") != 0)
		  continue;
	      break;

	    case (('_' << 16) | ('G' << 8) | 'O'):
	      if (strcmp (name, "_GOT_END_") != 0)
		  continue;
	      break;

	    case (('_' << 16) | ('S' << 8) | 'D'):
	      if (strcmp (name, "_SDA_BASE_") != 0)
		  continue;
	      break;

	    case (('_' << 16) | ('e' << 8) | 'n'):
	      if (strcmp (name, "_end") != 0)
		continue;
	      break;

	    case (('_' << 16) | ('e' << 8) | 'd'):
	      if (strcmp (name, "_edata") != 0)
		continue;
	      break;

	    case (('_' << 16) | ('e' << 8) | 't'):
	      if (strcmp (name, "_etext") != 0)
		continue;
	      break;

	    case (('e' << 16) | ('n' << 8) | 'd'):
	      if (strcmp (name, "end") != 0)
		continue;
	      break;

	    case (('e' << 16) | ('d' << 8) | 'a'):
	      if (strcmp (name, "edata") != 0)
		continue;
	      break;

	    case (('e' << 16) | ('t' << 8) | 'e'):
	      if (strcmp (name, "etext") != 0)
		continue;
	      break;
	    }

	  for (layout_ptr = vm_layout; layout_ptr != (layout_t *)0; layout_ptr = layout_ptr->next)
	    {
	      if (layout_ptr->name == name)
		break;

	      if (layout_ptr->name[0] == name[0] && strcmp (layout_ptr->name, name) == 0)
		break;
	    }

	  if (layout_ptr == (layout_t *)0)
	    add_layout (&vm_layout, name, sym.st_value, 0, layout_free_no);
	}
    }
}


int
main(int argc, char *argv[])
{
  int argnum;
  Elf32_Word i;
  int letter, j;
  int ch;
  const char *printable;
  size_t len, hdr_len, hdr2_len;
  time_t now;
  int base_machine;

  while ((letter = getopt (argc, argv, "dehln:pst:vx")) != EOF)
    {
      switch (letter)
	{
	default:
	  errors++;
	  break;

	case 'd':		/* invoke objdump to disassemble text sections */
	  objdump = 1;
	  break;

	case 'e':		/* put linker generated symbols in the VM layout */
	  linker_symbols = 1;
	  break;

	case 'h':		/* print raw ELF header */
	  print_header = 1;
	  break;

	case 'l':		/* skip printing layout information */
	  no_layout = 1;
	  break;

	case 'n':		/* print a specific section */
	  {
	    struct sec_list *p = (struct sec_list *) xmalloc (sizeof (struct sec_list));
	    p->next = sections;
	    p->section = optarg;
	    sections = p;
	    break;
	  }

	case 'p':		/* skip printing the progbits section body */
	  no_progbits = 1;
	  break;

	case 's':		/* print just the summary information */
	  summary = 1;
	  break;

	case 't':		/* specify output format */
	  format = ch = tolower (*optarg);
	  if (optarg[1] != '\0' || (ch != 'd' && ch != 'x' && ch != 'o' && ch != 'u' & ch != 'i'))
	    {
	      fprintf (stderr, "-t %s is not 'i', 'd', 'o', 'u', or 'x'\n", optarg);
	      errors++;
	    }
	  break;

	case 'v':		/* print version string */
	  print_vers = 1;
	  break;

	case 'x':		/* string table pointer is invalid */
	  ignore_string = 1;
	  break;
	}
    }

  if (errors || optind >= argc)
    {
      fprintf (stderr, "Calling sequence:\n\n");
      fprintf (stderr, "\telfdump [-n name]... [-delpsx] [-t format] files...\n");
      fprintf (stderr, "\nwhere:\n\n");
      fprintf (stderr, "-d\tInvoke objdump -d for executable sections.\n");
      fprintf (stderr, "-e\tPut the linker generated symbols in the vm layout.\n");
      fprintf (stderr, "-l\tDo not print file/VM layout information.\n");
      fprintf (stderr, "-n name\tOnly print section 'name'\n");
      fprintf (stderr, "-p\tDo not print the body of a progbits section.\n");
      fprintf (stderr, "-s\tOnly print section header summary information.\n");
      fprintf (stderr, "-t fmt\tSpecify output format for numbers (i, d, o, u, or x).\n");
      fprintf (stderr, "-v\tPrint version number.\n");
      fprintf (stderr, "-x\tAssume the string table is corrupt.\n");
      return 1;
    }

  if (print_vers)
    printf ("%s\n\n", elfdump_what_string);

  for (argnum = optind; (file = argv[argnum]) != (char *)0; argnum++)
    {
      if (argnum > optind)
	printf ("\f====================\n\n");

      if (alloc_file)
	free ((void *) file_bytes);

#ifndef NO_MMAP
      else if (file_bytes != (Uchar *)-1)
	munmap ((caddr_t)file_bytes, (size_t)file_stat.st_size);
#endif

      if (fixed_sect_hdrs)
	free ((void *) fixed_sect_hdrs);

      if (fixed_prog_hdrs)
	free ((void *) fixed_prog_hdrs);

      /* Map in the file.  */
      if (stat (file, &file_stat) < 0 || (file_fd = open (file, O_RDONLY)) < 0)
	{
	  perror (file);
	  errors++;
	  continue;
	}

#ifndef NO_MMAP
      if ((file_bytes = (Uchar *) mmap ((caddr_t)0,
					(size_t)file_stat.st_size,
					PROT_READ,
					MAP_PRIVATE | MAP_FILE | MAP_VARIABLE,
					file_fd,
					(off_t)0)) == (Uchar *)-1)
#endif
	{
	  file_bytes = (Uchar *) xmalloc ((size_t)file_stat.st_size);
	  if (read (file_fd, (void *) file_bytes, (size_t)file_stat.st_size) != file_stat.st_size)
	    {
	      perror (file);
	      errors++;
	      continue;
	    }
	}

      ehdr = *(Elf32_Ehdr *)file_bytes;
      if (file_stat.st_size < (off_t)sizeof (ehdr))
	{
	  fflush (stdout);
	  fprintf (stderr, "%s is too small to be an ELF file.\n", file);
	  fflush (stderr);
	  errors++;
	  continue;
	}

      if (ehdr.e_ident[EI_MAG0] != ELFMAG0
	  || ehdr.e_ident[EI_MAG1] != ELFMAG1
	  || ehdr.e_ident[EI_MAG2] != ELFMAG2
	  || ehdr.e_ident[EI_MAG3] != ELFMAG3)
	{
	  fflush (stdout);
	  fprintf (stderr, "%s is not an ELF file.\n", file);
	  fflush (stderr);
	  errors++;
	  continue;
	}

      /* Convert ehdr to host format */
      ehdr.e_type      = decode_elf32_half (ehdr.e_type);	/* object file type */
      ehdr.e_machine   = decode_elf32_half (ehdr.e_machine);	/* machine architecture */
      ehdr.e_version   = decode_elf32_word (ehdr.e_version);	/* object file version number */
      ehdr.e_entry     = decode_elf32_addr (ehdr.e_entry);	/* start address */
      ehdr.e_phoff     = decode_elf32_off (ehdr.e_phoff);	/* program header offset */
      ehdr.e_shoff     = decode_elf32_off (ehdr.e_shoff);	/* section table offset */
      ehdr.e_flags     = decode_elf32_word (ehdr.e_flags);	/* machine dependent flags */
      ehdr.e_ehsize    = decode_elf32_half (ehdr.e_ehsize);	/* elf header size */
      ehdr.e_phentsize = decode_elf32_half (ehdr.e_phentsize);	/* sizeof(Elf32_Phdr) */
      ehdr.e_phnum     = decode_elf32_half (ehdr.e_phnum);	/* # of program headers */
      ehdr.e_shentsize = decode_elf32_half (ehdr.e_shentsize);	/* sizeof(Elf32_Shdr) */
      ehdr.e_shnum     = decode_elf32_half (ehdr.e_shnum);	/* # of section headers */
      ehdr.e_shstrndx  = decode_elf32_half (ehdr.e_shstrndx);	/* section name table index */

      /* Print the raw header only if desired.  */
      if (print_header)
	{
	  hdr_len = stringify_len (ehdr.e_type, 'd', 1);

	  len = stringify_len (ehdr.e_machine, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_version, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_entry, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_phoff, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_shoff, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_flags, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_ehsize, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_phentsize, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_phnum, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_shentsize, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_shnum, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  len = stringify_len (ehdr.e_shstrndx, 'd', 1);
	  if (hdr_len < len)
	    hdr_len = len;

	  printf ("e_ident     = %s\n", printable_string ((Uchar *)ehdr.e_ident, EI_NIDENT));
	  printf ("e_type      = %s\n", stringify (ehdr.e_type,      hdr_len, 'd', 0, "", ""));
	  printf ("e_machine   = %s\n", stringify (ehdr.e_machine,   hdr_len, 'd', 0, "", ""));
	  printf ("e_version   = %s\n", stringify (ehdr.e_version,   hdr_len, 'd', 0, "", ""));
	  printf ("e_entry     = %s\n", stringify (ehdr.e_entry,     hdr_len, 'd', 0, "", ""));
	  printf ("e_phoff     = %s\n", stringify (ehdr.e_phoff,     hdr_len, 'd', 0, "", ""));
	  printf ("e_shoff     = %s\n", stringify (ehdr.e_shoff,     hdr_len, 'd', 0, "", ""));
	  printf ("e_flags     = %s\n", stringify (ehdr.e_flags,     hdr_len, 'd', 0, "", ""));
	  printf ("e_ehsize    = %s\n", stringify (ehdr.e_ehsize,    hdr_len, 'd', 0, "", ""));
	  printf ("e_phentsize = %s\n", stringify (ehdr.e_phentsize, hdr_len, 'd', 0, "", ""));
	  printf ("e_phnum     = %s\n", stringify (ehdr.e_phnum,     hdr_len, 'd', 0, "", ""));
	  printf ("e_shentsize = %s\n", stringify (ehdr.e_shentsize, hdr_len, 'd', 0, "", ""));
	  printf ("e_shnum     = %s\n", stringify (ehdr.e_shnum,     hdr_len, 'd', 0, "", ""));
	  printf ("e_shstrndx  = %s\n", stringify (ehdr.e_shstrndx,  hdr_len, 'd', 0, "", ""));
	  continue;
	}

      if (ehdr.e_ehsize != sizeof (ehdr))
	{
	  fflush (stdout);
	  fprintf (stderr, "%s elf header is %s bytes, expected %s\n", file,
		   stringify (ehdr.e_ehsize, 0, 'd', 0, "", ""),
		   stringify (sizeof (ehdr), 0, 'd', 0, "", ""));

	  fflush (stderr);
	  errors++;
	  continue;
	}

      /* Convert the section headers to host format */
      if (ehdr.e_shnum == 0)
	fixed_sect_hdrs = (Elf32_Shdr *)0;

      else
	{
	  fixed_sect_hdrs = xcalloc (sizeof (Elf32_Shdr), ehdr.e_shnum);
	  for (i = 0; i < ehdr.e_shnum; i++)
	    decode_elf32_shdr (&fixed_sect_hdrs[i], &((Elf32_Shdr *)(file_bytes + ehdr.e_shoff))[i]);
	}

      /* Convert the program headers to host format */
      if (ehdr.e_phnum == 0)
	fixed_prog_hdrs = (Elf32_Phdr *)0;

      else
	{
	  fixed_prog_hdrs = xcalloc (sizeof (Elf32_Phdr), ehdr.e_phnum);
	  for (i = 0; i < ehdr.e_phnum; i++)
	    decode_elf32_phdr (&fixed_prog_hdrs[i], &((Elf32_Phdr *)(file_bytes + ehdr.e_phoff))[i]);
	}

      add_layout (&file_layout, "Global header", 0, sizeof (ehdr), layout_free_no);

      if (ehdr.e_shnum)
	add_layout (&file_layout, "Section headers", ehdr.e_shoff, ehdr.e_shnum * ehdr.e_shentsize, layout_free_no);

      if (ehdr.e_phnum)
	add_layout (&file_layout, "Program headers", ehdr.e_phoff, ehdr.e_phnum * ehdr.e_phentsize, layout_free_no);

      now = time ((time_t *)0);
      printf ("File: %s, modified %s\n", file, ctime (&now));

      /* Size and print the header fields */
      switch (ehdr.e_type)
	{
	default:	printable = "";			break;
	case ET_NONE:	printable = ", no file type";	break;
	case ET_REL:	printable = ", relocatable";	break;
	case ET_EXEC:	printable = ", executable";	break;
	case ET_DYN:	printable = ", shared object";	break;
	case ET_CORE:	printable = ", core file";	break;
	}

      hdr_field[ (int)HDR_TYPE ].print  = printable;
      hdr_field[ (int)HDR_TYPE ].title  = "type:";
      hdr_field[ (int)HDR_TYPE ].value  = ehdr.e_type;
      hdr_field[ (int)HDR_TYPE ].format = 'd';

      switch (ehdr.e_machine)
	{
	default:		  printable = "";				break;
	case EM_NONE:		  printable = ", no machine";			break;
	case EM_M32:		  printable = ", we 32100";			break;
	case EM_SPARC:		  printable = ", sparc";			break;
	case EM_386:		  printable = ", i386";				break;
	case EM_68K:		  printable = ", m68k";				break;
	case EM_88K:		  printable = ", m88k";				break;
	case EM_860:		  printable = ", i860";				break;
	case EM_MIPS:		  printable = ", mips";				break;
	case EM_MIPS_RS4_BE:	  printable = ", mips r4000";			break;
	case EM_SPARC64:	  printable = ", sparc 64 bit";			break;
	case EM_PARISC:		  printable = ", pa-risc";			break;
	case EM_PPC:		  printable = ", powerpc";			break;
	case EM_PPC_OLD:	  printable = ", old powerpc";			break;
	case EM_CYGNUS_POWERPC:	  printable = ", cygnus powerpc";		break;
	case EM_CYGNUS_ARC:	  printable = ", cygnus arc";			break;
		
	case EM_M32	    << 8: printable = ", we 32100 (swapped)";		break;
	case EM_SPARC	    << 8: printable = ", sparc (swapped)";		break;
	case EM_386	    << 8: printable = ", i386 (swapped)";		break;
	case EM_68K	    << 8: printable = ", m68k (swapped)";		break;
	case EM_88K	    << 8: printable = ", m88k (swapped)";		break;
	case EM_860	    << 8: printable = ", i860 (swapped)";		break;
	case EM_MIPS	    << 8: printable = ", mips (swapped)";		break;
	case EM_MIPS_RS4_BE << 8: printable = ", mips r4000 (swapped)";		break;
	case EM_SPARC64     << 8: printable = ", sparc 64 bit (swapped)";	break;
	case EM_PARISC      << 8: printable = ", pa-risc (swapped)";		break;
	case EM_PPC         << 8: printable = ", powerpc (swapped)";		break;
	case EM_PPC_OLD     << 8: printable = ", old powerpc (swapped)";	break;
	}

      hdr_field[ (int)HDR_MACHINE ].print  = printable;
      hdr_field[ (int)HDR_MACHINE ].title  = "machine:";
      hdr_field[ (int)HDR_MACHINE ].value  = ehdr.e_machine;
      hdr_field[ (int)HDR_MACHINE ].format = 'd';

      switch (ehdr.e_ident[EI_CLASS])
	{
	default:	 printable = "";		break;
	case ELFCLASS32: printable = ", 32-bit";	break;
	case ELFCLASS64: printable = ", 64-bit";	break;
	}

      hdr_field[ (int)HDR_CLASS ].print  = printable;
      hdr_field[ (int)HDR_CLASS ].title  = "class:";
      hdr_field[ (int)HDR_CLASS ].value  = ehdr.e_ident[EI_CLASS];
      hdr_field[ (int)HDR_CLASS ].format = 'd';

      switch (ehdr.e_ident[EI_DATA])
	{
	default:	  printable = "";		 break;
	case ELFDATA2LSB: printable = ", little endian"; break;
	case ELFDATA2MSB: printable = ", big endian";	 break;
	}

      hdr_field[ (int)HDR_DATA ].print  = printable;
      hdr_field[ (int)HDR_DATA ].title  = "data:";
      hdr_field[ (int)HDR_DATA ].value  = ehdr.e_ident[EI_DATA];
      hdr_field[ (int)HDR_DATA ].format = 'd';

      hdr_field[ (int)HDR_VERSION ].print  = "";
      hdr_field[ (int)HDR_VERSION ].title  = "version:";
      hdr_field[ (int)HDR_VERSION ].value  = ehdr.e_version;
      hdr_field[ (int)HDR_VERSION ].format = 'd';

      flags_print[0] = '\0';
      switch (ehdr.e_machine)
	{
	default:		base_machine = ehdr.e_machine;	break;
	case EM_PPC_OLD:	base_machine = EM_PPC;		break;
	case EM_CYGNUS_POWERPC:	base_machine = EM_PPC;		break;
	case EM_MIPS_RS4_BE:	base_machine = EM_MIPS;		break;
	}

      if (ehdr.e_ident[EI_CLASS] == ELFCLASS32 && ehdr.e_machine != 0)
	{
	  for (j = 0; elf32_flags[j].string != (const char *)0; j++)
	    {
	      if (base_machine == elf32_flags[j].machine
		  && (ehdr.e_flags & elf32_flags[j].mask) == elf32_flags[j].value)
		{
		  strcat (flags_print, elf32_flags[j].string);
		}
	    }
	}

      hdr_field[ (int)HDR_FLAGS ].print  = flags_print;
      hdr_field[ (int)HDR_FLAGS ].title  = "flags:";
      hdr_field[ (int)HDR_FLAGS ].value  = ehdr.e_flags;
      hdr_field[ (int)HDR_FLAGS ].format = 'x';

      hdr_field[ (int)HDR_ENTRY ].print  = "";
      hdr_field[ (int)HDR_ENTRY ].title  = "entry:";
      hdr_field[ (int)HDR_ENTRY ].value  = ehdr.e_entry;
      hdr_field[ (int)HDR_ENTRY ].format = 'x';

      hdr_field[ (int)HDR_EHSIZE ].print  = "";
      hdr_field[ (int)HDR_EHSIZE ].title  = "header size:";
      hdr_field[ (int)HDR_EHSIZE ].value  = ehdr.e_ehsize;
      hdr_field[ (int)HDR_EHSIZE ].format = 'd';

      hdr_field[ (int)HDR_SHSTRNDX ].print  = "";
      hdr_field[ (int)HDR_SHSTRNDX ].title  = "section string index:";
      hdr_field[ (int)HDR_SHSTRNDX ].value  = ehdr.e_shstrndx;
      hdr_field[ (int)HDR_SHSTRNDX ].format = 'd';

      hdr_len = hdr2_len = 0;
      for (j = 0; j < (int)HDR_MAX; j++)
	{
	  if (!hdr_field[j].title)
	    continue;

	  len = strlen (hdr_field[j].title);
	  if (len > hdr_len)
	    hdr_len = len;

	  len = stringify_len (hdr_field[j].value, hdr_field[j].format, 1);
	  if (len > hdr2_len)
	    hdr2_len = len;
	}

      for (j = 0; j < (int)HDR_MAX; j++)
	{
	  if (!hdr_field[j].title)
	    continue;

	  printf ("    %-*s %s%s%s\n",
		  (int)hdr_len, hdr_field[j].title,
		  (format || hdr_field[j].format == 'x') ? "" : "  ",
		  stringify (hdr_field[j].value, hdr2_len, hdr_field[j].format, 0, "", ""),
		  hdr_field[j].print);
	}

      printf ("\n");
      printf ("# sections = %s, section element size = %s, section offset = %s\n",
	      stringify (ehdr.e_shnum, 2, 'd', 0, "", ""),
	      stringify (ehdr.e_shentsize, 2, 'd', 0, "", ""),
	      stringify (ehdr.e_shoff, 6, 'd', 0, "", ""));

      printf ("# headers  = %s, header  element size = %s, header  offset = %s\n\n",
	      stringify (ehdr.e_phnum, 2, 'd', 0, "", ""),
	      stringify (ehdr.e_phentsize, 2, 'd', 0, "", ""),
	      stringify (ehdr.e_phoff, 6, 'd', 0, "", ""));

      /* Print each of the sections now.  */
      for (i = 0; i < ehdr.e_shnum; i++)
	size_section (i);

      for (i = 0; i < ehdr.e_shnum; i++)
	print_section (i);

      /* Print each of the program headers now.  */
      if (ehdr.e_phnum != 0)
	print_proghdr ();

      /* Print the layout information if desired */
      if (!no_layout)
	{
	  print_layout (file_layout, "ELF file");

	  if (ehdr.e_phnum)
	    {
	      if (linker_symbols)
		find_linker_symbols ();

	      print_layout (vm_layout, "Virtual Memory");
	    }
	}

      if (alloc_file)
	free ((void *) file_bytes);

#ifndef NO_MMAP
      else if (file_bytes != (Uchar *)-1)
	munmap ((caddr_t)file_bytes, (size_t)file_stat.st_size);
#endif

      if (file_fd >= 0)
	close (file_fd);

      if (file_layout)
	free_layout (&file_layout);

      if (vm_layout)
	free_layout (&vm_layout);

      if (fixed_sect_hdrs)
	free ((void *) fixed_sect_hdrs);

      if (fixed_prog_hdrs)
	free ((void *) fixed_prog_hdrs);

      file_bytes = (Uchar *)-1;
      file_fd = -1;
      fixed_sect_hdrs = (Elf32_Shdr *)0;
      fixed_prog_hdrs = (Elf32_Phdr *)0;
    }

  return (errors) ? 1 : 0;
}
